import sys
import os
import atexit
import logging
import socket
from django.utils.translation import ugettext as _
#bots-modules
import botslib
import botsinit
import botsglobal


def sqlite_database_is_version3():
    for row in botslib.query('''PRAGMA table_info(routes)'''):
        if row['name'] == 'translateind':
            if row['type'] == 'bool':
                return False
            else:
                return True
    raise Exception('Could not determine version of database')


QUERYSTRING = '''
PRAGMA writable_schema = 1;
UPDATE SQLITE_MASTER SET SQL = 
'CREATE TABLE "routes" (
    "id" integer NOT NULL PRIMARY KEY,
    "idroute" varchar(35) NOT NULL,
    "seq" integer unsigned NOT NULL,
    "active" bool NOT NULL,
    "fromchannel_id" varchar(35) REFERENCES "channel" ("idchannel"),
    "fromeditype" varchar(35) NOT NULL,
    "frommessagetype" varchar(35) NOT NULL,
    "tochannel_id" varchar(35) REFERENCES "channel" ("idchannel"),
    "toeditype" varchar(35) NOT NULL,
    "tomessagetype" varchar(35) NOT NULL,
    "alt" varchar(35) NOT NULL,
    "frompartner_id" varchar(35) REFERENCES "partner" ("idpartner"),
    "topartner_id" varchar(35) REFERENCES "partner" ("idpartner"),
    "frompartner_tochannel_id" varchar(35) REFERENCES "partner" ("idpartner"),
    "topartner_tochannel_id" varchar(35) REFERENCES "partner" ("idpartner"),
    "testindicator" varchar(1) NOT NULL,
    "translateind" integer NOT NULL,
    "notindefaultrun" bool NOT NULL,
    "desc" text,
    "rsrv1" varchar(35),
    "rsrv2" integer,
    "defer" bool NOT NULL,
    "zip_incoming" integer,
    "zip_outgoing" integer,
    UNIQUE ("idroute", "seq"))' 
WHERE NAME = 'routes';
PRAGMA writable_schema = 0;
'''

def sqlite3():
    if sqlite_database_is_version3():
        print 'Database sqlite3 is already bots version 3. No action is taken.'
        return 2

    print 'Start changing sqlite3 database to bots version 3.'
    cursor = botsglobal.db.cursor()
    try:
        #channel ****************************************
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "rsrv3" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "keyfile" VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "certfile" VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "testpath" VARCHAR(256) ''')
        #filereport ****************************************
        cursor.execute('''DROP INDEX "filereport_reportidta" ''')
        cursor.execute('''ALTER TABLE "filereport" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        #partner *************************************
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr1" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr2" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr3" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr4" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr5" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name1" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name2" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name3" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address1" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address2" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address3" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "city" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "postalcode" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "countrysubdivision" VARCHAR(9) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "countrycode" VARCHAR(3) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "phone1" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "phone2" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "desc" TEXT ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "startdate" DATE ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "enddate" DATE ''')
        #report ****************************************
        cursor.execute('''CREATE INDEX "report_ts" ON "report" ("ts")''')
        cursor.execute('''ALTER TABLE "report" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "report" ADD COLUMN "acceptance" INTEGER DEFAULT 0''')
        #routes ****************************************
        cursor.execute('''ALTER TABLE "routes" ADD COLUMN "zip_incoming" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "routes" ADD COLUMN "zip_outgoing" INTEGER DEFAULT 0''')
        #ta ****************************************
        cursor.execute('''DROP INDEX "ta_script" ''')
        cursor.execute('''CREATE INDEX "ta_reference" ON "ta" ("reference")''')
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "numberofresends" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "rsrv5" VARCHAR(35) DEFAULT '' ''')
    except:
        txt = botslib.txtexc()
        botsglobal.db.rollback()
        cursor.close()
        print 'Error in adding fields to sqlite3 database: "%s".'%(txt)
        return 1
    else:
        botsglobal.db.commit()
        cursor.close()

    cursor = botsglobal.db.cursor()
    try:
        cursor.executescript(QUERYSTRING)
    except:
        txt = botslib.txtexc()
        botsglobal.db.rollback()
        cursor.close()
        print 'Error in changing sqlite3 database-schema "routes": "%s".'%(txt)
        return 1
    else:
        botsglobal.db.commit()
        cursor.close()
        
    print 'Succesful changed sqlite3 database to bots version 3.'
    return 0
            


def postgresql_psycopg2():
    print 'Start changing postgresql database to bots version 3.'
    cursor = botsglobal.db.cursor()
    try:
        #channel ****************************************
        cursor.execute('''ALTER TABLE "channel" ALTER COLUMN "filename" TYPE VARCHAR(256)''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "rsrv3" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "keyfile" VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "certfile" VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE "channel" ADD COLUMN "testpath" VARCHAR(256) ''')
        #ccode ****************************************
        cursor.execute('''ALTER TABLE "ccode" ALTER COLUMN "rightcode" TYPE VARCHAR(70)''')
        cursor.execute('''ALTER TABLE "ccode" ALTER COLUMN "attr1" TYPE VARCHAR(70)''')
        #filereport ****************************************
        cursor.execute('''ALTER TABLE "filereport" DROP CONSTRAINT "filereport_pkey" ''')      #remove primary key
        cursor.execute('''ALTER TABLE "filereport" DROP CONSTRAINT "filereport_idta_key" ''')  #drop contraint UNIQUE(idta, reportidta)
        cursor.execute('''DROP INDEX "filereport_idta" ''')                                  #drop index on idta (will be primary key)
        cursor.execute('''DROP INDEX "filereport_reportidta" ''')
        cursor.execute('''ALTER TABLE "filereport" DROP COLUMN "id" ''')     
        cursor.execute('''ALTER TABLE "filereport" ADD CONSTRAINT "filereport_pkey" PRIMARY KEY("idta")''')    #idta is primary key
        cursor.execute('''ALTER TABLE "filereport" ALTER COLUMN "errortext" TYPE TEXT''')
        cursor.execute('''ALTER TABLE "filereport" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "filereport" ADD COLUMN "acceptance" INTEGER DEFAULT 0''')
        #partner *************************************
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr1" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr2" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr3" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr4" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "attr5" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name1" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name2" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "name3" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address1" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address2" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "address3" VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "city" VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "postalcode" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "countrysubdivision" VARCHAR(9) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "countrycode" VARCHAR(3) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "phone1" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "phone2" VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "desc" TEXT ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "startdate" DATE ''')
        cursor.execute('''ALTER TABLE "partner" ADD COLUMN "enddate" DATE ''')
        #persist ****************************************
        cursor.execute('''ALTER TABLE "persist" ALTER COLUMN "content" TYPE TEXT''')
        #report ****************************************
        cursor.execute('''CREATE INDEX "report_ts" ON "report" ("ts")''')
        cursor.execute('''ALTER TABLE "report" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        #routes ****************************************
        cursor.execute('''ALTER TABLE "routes" ALTER COLUMN "translateind" TYPE integer USING CASE WHEN "translateind"=FALSE THEN 0 ELSE 1 END''')
        cursor.execute('''ALTER TABLE "routes" ADD COLUMN "zip_incoming" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "routes" ADD COLUMN "zip_outgoing" INTEGER DEFAULT 0''')
        #ta ****************************************
        cursor.execute('''DROP INDEX "ta_script" ''')
        cursor.execute('''CREATE INDEX "ta_reference" ON "ta" ("reference")''')
        cursor.execute('''ALTER TABLE "ta" ALTER COLUMN "errortext" TYPE TEXT ''') 
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "filesize" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "numberofresends" INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE "ta" ADD COLUMN "rsrv5" VARCHAR(35) DEFAULT '' ''')
    except:
        txt = botslib.txtexc()
        botsglobal.db.rollback()
        cursor.close()
        print 'Error in changing postgresql database: "%s".'%(txt)
        return 1
    else:
        botsglobal.db.commit()
        cursor.close()
        
    print 'Succesful changed postgresql database to bots version 3.'
    return 0


def mysql():
    print 'Start changing mysql database to bots version 3.'
    cursor = botsglobal.db.cursor()
    try:
        #channel ****************************************
        cursor.execute('''ALTER TABLE `channel` MODIFY `filename` VARCHAR(256) NOT NULL''')
        cursor.execute('''ALTER TABLE `channel` ADD COLUMN `rsrv3` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `channel` ADD COLUMN `keyfile` VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE `channel` ADD COLUMN `certfile` VARCHAR(256) ''')
        cursor.execute('''ALTER TABLE `channel` ADD COLUMN `testpath` VARCHAR(256) ''')
        #ccode ****************************************
        cursor.execute('''ALTER TABLE `ccode` MODIFY `rightcode` VARCHAR(70)''')
        cursor.execute('''ALTER TABLE `ccode` MODIFY `attr1` VARCHAR(70)''')
        #filereport ****************************************
        cursor.execute('''ALTER TABLE `filereport` CHANGE `id` `id` INTEGER ''')    #drop autoincrement
        cursor.execute('''ALTER TABLE `filereport` DROP PRIMARY KEY ''')            #drop index on id
        cursor.execute('''ALTER TABLE `filereport` DROP COLUMN `id` ''')            #drop id veld
        cursor.execute('''ALTER TABLE `filereport` DROP KEY `idta` ''')             #remove UNIQUE constraint     
        #~ cursor.execute('''ALTER TABLE `filereport` DROP INDEX `reportidta` ''')  #not possible as index name is not known
        cursor.execute('''ALTER TABLE `filereport` MODIFY `errortext` TEXT''')
        cursor.execute('''ALTER TABLE `filereport` ADD COLUMN `filesize` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `filereport` ADD COLUMN `acceptance` INTEGER DEFAULT 0''')
        #partner *************************************
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `attr1` VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `attr2` VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `attr3` VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `attr4` VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `attr5` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `name1` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `name2` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `name3` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `address1` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `address2` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `address3` VARCHAR(70) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `city` VARCHAR(35) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `postalcode` VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `countrysubdivision` VARCHAR(9) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `countrycode` VARCHAR(3) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `phone1` VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `phone2` VARCHAR(17) ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `desc` TEXT ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `startdate` DATE ''')
        cursor.execute('''ALTER TABLE `partner` ADD COLUMN `enddate` DATE ''')
        #persist ****************************************
        cursor.execute('''ALTER TABLE `persist` MODIFY `content` TEXT''')
        #report ****************************************
        cursor.execute('''CREATE INDEX `report_ts` ON `report` (`ts`)''')
        cursor.execute('''ALTER TABLE `report` ADD COLUMN `filesize` INTEGER DEFAULT 0''')
        #routes ****************************************
        cursor.execute('''ALTER TABLE `routes` ADD COLUMN `zip_incoming` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `routes` ADD COLUMN `zip_outgoing` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `routes` MODIFY `translateind` integer NOT NULL''')
        #ta ****************************************
        #~ cursor.execute('''ALTER TABLE `ta` DROP INDEX `script` ''')   #not possible as index name is not known
        cursor.execute('''CREATE INDEX `ta_reference` ON `ta` (`reference`)''') 
        cursor.execute('''ALTER TABLE `ta` ADD COLUMN `filesize` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `ta` ADD COLUMN `numberofresends` INTEGER DEFAULT 0''')
        cursor.execute('''ALTER TABLE `ta` ADD COLUMN `rsrv5` VARCHAR(35) DEFAULT '' ''')
        cursor.execute('''ALTER TABLE `ta` MODIFY `errortext` TEXT ''')
    except:
        txt = botslib.txtexc()
        botsglobal.db.rollback()
        cursor.close()
        print 'Error in changing mysql database: "%s".'%(txt)
        return 1
    else:
        botsglobal.db.commit()
        cursor.close()
        
    print 'Succesful changed mysql database to bots version 3.'
    return 0


def start():
    #********command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    Updates existing bots database to version %(version)s

    Usage:
        %(name)s  [config-option]
    Options:
        -c<directory>        directory for configuration files (default: config).

    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(3)
        else:   #pick up names of routes to run
            print usage
            sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.

    #**************check if another instance of bots-engine is running/if port is free******************************
    try:
        engine_socket = botslib.check_if_other_engine_is_running()
    except socket.error:
        sys.exit(3)
    else:
        atexit.register(engine_socket.close)

    #**************initialise logging******************************
    process_name = 'updatedatabase'
    botsglobal.logger = botsinit.initenginelogging(process_name)
    atexit.register(logging.shutdown)
    for key,value in botslib.botsinfo():    #log info about environement, versions, etc
        botsglobal.logger.info(u'%(key)s: "%(value)s".',{'key':key,'value':value})

    #**************connect to database**********************************
    try:
        botsinit.connect()
    except Exception as msg:
        botsglobal.logger.exception(_(u'Could not connect to database. Database settings are in bots/config/settings.py. Error: "%(msg)s".'),{'msg':msg})
        sys.exit(3)
    else:
        botsglobal.logger.info(_(u'Connected to database.'))
        atexit.register(botsglobal.db.close)

    #**************handle database lock****************************************
    #set a lock on the database; if not possible, the database is locked: an earlier instance of bots-engine was terminated unexpectedly.
    if not botslib.set_database_lock():
        warn =  _(u'!Bots database is locked!\n'\
                    'Bots-engine has ended in an unexpected way during the last run.\n'\
                    'Most likely causes: sudden power-down, system crash, problems with disk I/O, bots-engine terminated by user, etc.')
        botsglobal.logger.critical(warn)
        sys.exit(3)
    atexit.register(botslib.remove_database_lock)

    if botsglobal.settings.DATABASES['default']['ENGINE'] == 'django.db.backends.sqlite3':
        terug = sqlite3()
    elif botsglobal.settings.DATABASES['default']['ENGINE'] == 'django.db.backends.mysql':
        terug = mysql()
    elif botsglobal.settings.DATABASES['default']['ENGINE'] == 'django.db.backends.postgresql_psycopg2':
        terug = postgresql_psycopg2()
    
    sys.exit(terug)


if __name__ == '__main__':
    start()




#!/usr/bin/env python
from bots import grammarcheck

if __name__ == '__main__':
    grammarcheck.start()
    #~ grammarcheck.startmulti('bots/usersys/grammars/edifact/*','edifact')     #for bulk check of grammars



import sys
import copy
import datetime
import re
import django
from django.core.paginator import Paginator, EmptyPage, InvalidPage
from django.utils.translation import ugettext as _
import models
import botsglobal
from botsconfig import *

def save_int(value):
    try:
        return int(value)
    except:
        return 0

def preparereport2view(post,runidta):
    terugpost = post.copy()
    thisrun = models.report.objects.get(idta=runidta)
    terugpost['datefrom'] = thisrun.ts
    try:
        nextrun = thisrun.get_next_by_ts()
        terugpost['dateuntil'] = nextrun.ts
    except:
        terugpost['dateuntil'] = datetimeuntil()
    return terugpost

def changepostparameters(post,soort):
    terugpost = post.copy()
    if soort == 'confirm2in':
        for key in ['confirmtype','confirmed','fromchannel','tochannel']:
            terugpost.pop(key)
        terugpost['ineditype'] = terugpost.pop('editype')[0]
        terugpost['inmessagetype'] = terugpost.pop('messagetype')[0]
        #~ terugpost['outeditype'] = ''
        #~ terugpost['outmessagetype'] = ''
    elif soort == 'confirm2out':
        for key in ['confirmtype','confirmed','fromchannel','tochannel']:
            terugpost.pop(key)
    elif soort == 'out2in':
        terugpost['outeditype'] = terugpost.pop('editype')[0]
        terugpost['outmessagetype'] = terugpost.pop('messagetype')[0]
        #~ terugpost['ineditype'] = ''
        #~ terugpost['inmessagetype'] = ''
    elif soort == 'out2confirm':
        for key in ['lastrun']:
            terugpost.pop(key)
    elif soort == 'in2out':
        terugpost['editype'] = terugpost.pop('outeditype')[0]
        terugpost['messagetype'] = terugpost.pop('outmessagetype')[0]
        for key in ['ineditype','inmessagetype']:
            terugpost.pop(key)
    elif soort == 'in2confirm':
        terugpost['editype'] = terugpost.pop('outeditype')[0]
        terugpost['messagetype'] = terugpost.pop('outmessagetype')[0]
        for key in ['lastrun','statust','ineditype','inmessagetype']:
            terugpost.pop(key)
    elif soort == '2process':
        #when going to process, most parameters are deleted.
        for key in terugpost.keys():
            if key not in ['datefrom','dateuntil','lastrun','idroute']: #keep these
                terugpost.pop(key)
    elif soort == 'fromprocess':
        pass    #is OK, all values are used
    terugpost['sortedby'] = 'ts'
    terugpost['sortedasc'] = False
    terugpost['page'] = 1
    return terugpost

def django_trace_origin(idta,where):
    ''' bots traces back all from the current step/ta_object.
        where is a dict that is used to indicate a condition.
        eg:  {'status':EXTERNIN}
        If bots finds a ta_object for which this is true, the ta_object is added to a list.
        The list is returned when all tracing is done, and contains all ta_object's for which 'where' is True
    '''
    def trace_recurse(ta_object):
        ''' recursive
            walk over ta_object's backward (to origin).
            if condition is met, add the ta_object to a list
        '''
        for parent in get_parent(ta_object):
            donelijst.append(parent.idta)
            for key,value in where.iteritems():
                if getattr(parent,key) != value:
                    break
            else:   #all where-criteria are true; check if we already have this ta_object
                teruglijst.append(parent)
            trace_recurse(parent)
    def get_parent(ta_object):
        ''' yields the parents of a ta_object '''
        if ta_object.parent:
            if ta_object.parent not in donelijst:   #search via parent
                yield models.ta.objects.get(idta=ta_object.parent)
        else:
            for parent in models.ta.objects.filter(idta__range=(ta_object.script,ta_object.idta),child=ta_object.idta):
                if parent.idta in donelijst:
                    continue
                yield parent

    donelijst = []
    teruglijst = []
    ta_object = models.ta.objects.get(idta=idta)
    trace_recurse(ta_object)
    return teruglijst


def trace_document(pquery):
    ''' trace forward & backwardfrom the current step/ta_object (status SPLITUP).
        gathers confirm information
    '''
    def trace_forward(ta_object):
        ''' recursive. walk over ta_object's forward (to exit). '''
        if ta_object.child:
            child = models.ta.objects.get(idta=ta_object.child)
        else:
            try:
                child = models.ta.objects.filter(parent=ta_object.idta)[0]
            except IndexError:
                return    #no result, return
        if child.confirmasked:
            taorg.confirmtext += u'%(confirmtype)s\n' % {'confirmasked':child.confirmasked,'confirmed':child.confirmed,'confirmtype':child.confirmtype}
            taorg.confirmidta = child.confirmidta
        if child.status == EXTERNOUT:
            taorg.outgoing = child.idta
            taorg.channel = child.tochannel
        trace_forward(child)
    def trace_back(ta_object):
        ''' recursive. walk over ta_object's backward (to origin).  '''
        if ta_object.parent:
            parent = models.ta.objects.get(idta=ta_object.parent)
        else:
            try:
                parent = models.ta.objects.filter(idta__range=(ta_object.script,ta_object.idta),child=ta_object.idta)[0]   #just get one parent
            except IndexError:
                return    #no result, return
        if parent.confirmasked:
            taorg.confirmtext += u'%(confirmtype)s\n' % {'confirmasked':parent.confirmasked,'confirmed':parent.confirmed,'confirmtype':parent.confirmtype}
            taorg.confirmidta = parent.confirmidta
        if parent.status == EXTERNIN:
            taorg.incoming = parent.idta
            taorg.channel = parent.fromchannel
        trace_back(parent)
    #main for trace_document*****************
    for taorg in pquery.object_list:
        taorg.confirmtext = u''
        if taorg.status == SPLITUP:
            trace_back(taorg)
        else:
            trace_forward(taorg)


def gettrace(ta_object):
    ''' recursive. Builds a tree of ta's (a trace) for parameter ta_object.
        children are a list in ta.
    '''
    if ta_object.child:  #has a explicit child
        ta_object.talijst = [models.ta.objects.get(idta=ta_object.child)]
    else:   #search in ta_object-table who is reffering to ta_object
        ta_object.talijst = list(models.ta.objects.filter(parent=ta_object.idta))
    for child in ta_object.talijst:
        gettrace(child)

def delete_from_ta(ta_object):
    ''' try to delete in ta table as much as possible.
        until a MERGE, this is easy.
        but a MERGE can contain messages from other infiles...
        in that case, just leave the MERGE 
    '''
    def gather_tas_before_merge(ta_object):
        ''' loop over ta tree untill MERGE.
            put all ta status MERGE in one list, others in deletelist
        '''
        if ta_object.status == MERGED:
            tas_merge.append(ta_object)
            return
        else:
            tas_for_deletion.append(ta_object)
        if ta_object.child:  #has a explicit child
            tmp_list = [models.ta.objects.get(idta=ta_object.child)]
        else:   #search in ta_object-table who is reffering to ta_object
            tmp_list = list(models.ta.objects.filter(parent=ta_object.idta))
        for child in tmp_list:
            gather_tas_before_merge(child)
    def gather_tas_after_merge(ta_object):
        tas_for_deletion.append(ta_object)
        if ta_object.child:  #has a explicit child
            tmp_list = [models.ta.objects.get(idta=ta_object.child)]
        else:   #search in ta_object-table who is reffering to ta_object
            tmp_list = list(models.ta.objects.filter(parent=ta_object.idta))
        for child in tmp_list:
            gather_tas_after_merge(child)
    tas_for_deletion = []
    tas_merge = []
    gather_tas_before_merge(ta_object)
    tas_merge = list(set(tas_merge))        #one MERGE only once in list
    for ta_merge in tas_merge:  #for each MERGE ta:
        #if all included files in the MERGED file are to be deleted, MERGED can be deleted 
        for includedta in models.ta.objects.filter(idta__range=(ta_merge.script,ta_merge.idta),child=ta_merge.idta):    #select all db-ta_object's included in MERGED ta_object
            if includedta not in tas_for_deletion:
                break
        else:
            gather_tas_after_merge(ta_merge)
    for ta_object in tas_for_deletion:
        ta_object.delete()

def trace2detail(ta_object):
    def newbranche(ta_object,level=0):
        def dota(ta_object, isfirststep = False):
            levelindicator = (level)*'| '
            if isfirststep and level:
                ta_object.ind = levelindicator[:-2] + '___'
            elif ta_object.status == FILEOUT and ta_object.nrmessages > 1:
                ta_object.ind = levelindicator
            elif ta_object.status == EXTERNOUT:
                if levelindicator:
                    ta_object.ind = levelindicator[:-2] + 'o=='
                else:
                    ta_object.ind = levelindicator[:-2]
            else:
                ta_object.ind = levelindicator
            #~ ta_object.action = models.ta.objects.only('filename').get(idta=ta_object.script)
            ta_object.channel = ta_object.fromchannel
            if ta_object.tochannel:
                ta_object.channel = ta_object.tochannel
            detaillist.append(ta_object)
            lengtetalijst = len(ta_object.talijst)
            if lengtetalijst > 1:
                for child in ta_object.talijst:
                    newbranche(child,level=level+1)
            elif lengtetalijst == 1:
                dota(ta_object.talijst[0])
        #start new level
        dota(ta_object,isfirststep = True)
    detaillist = []
    newbranche(ta_object)
    return detaillist

def datetimefrom():
    terug = datetime.datetime.today() - datetime.timedelta(days=botsglobal.ini.getint('settings','maxdays',30))
    return terug.strftime('%Y-%m-%d 00:00:00')

def datetimeuntil():
    terug = datetime.datetime.today()
    return terug.strftime('%Y-%m-%d 23:59:59')

def handlepagination(requestpost,cleaned_data):
    ''' use requestpost to set criteria for pagination in cleaned_data'''
    if "first" in requestpost:
        cleaned_data['page'] = 1
    elif "previous" in requestpost:
        cleaned_data['page'] = cleaned_data['page'] - 1
    elif "next" in requestpost:
        cleaned_data['page'] = cleaned_data['page'] + 1
    elif "last" in requestpost:
        cleaned_data['page'] = sys.maxint
    elif "order" in requestpost:   #change the sorting order
        if requestpost['order'] == cleaned_data['sortedby']:  #sort same row, but desc->asc etc
            cleaned_data['sortedasc'] =  not cleaned_data['sortedasc']
        else:
            cleaned_data['sortedby'] = requestpost['order'].lower()
            if cleaned_data['sortedby'] == 'ts':
                cleaned_data['sortedasc'] = False
            else:
                cleaned_data['sortedasc'] = True

def getidtalastrun():
    return models.filereport.objects.all().aggregate(django.db.models.Max('reportidta'))['reportidta__max']

def filterquery(query , org_cleaned_data, incoming=False, paginate=True):
    ''' filter query using the data of the form (mostly in hidden fields).
        parameter 'paginate' controls if pagination is used or not.
    '''
    cleaned_data = copy.copy(org_cleaned_data)  #copy because it it destroyed in setting up query
    page = cleaned_data.pop('page')             #do not use this in query, use in paginator
    if 'dateuntil' in cleaned_data:
        query = query.filter(ts__lt=cleaned_data.pop('dateuntil'))
    if 'datefrom' in cleaned_data:
        query = query.filter(ts__gte=cleaned_data.pop('datefrom'))
    if 'sortedby' in cleaned_data:
        query = query.order_by({True:'',False:'-'}[cleaned_data.pop('sortedasc')] + cleaned_data.pop('sortedby'))
    if 'lastrun' in cleaned_data:
        if cleaned_data.pop('lastrun'):
            idtalastrun = getidtalastrun()
            if idtalastrun:     #if no result (=None): there are no filereports.
                if incoming:    #detect if incoming; do other selection
                    query = query.filter(reportidta=idtalastrun)
                else:
                    query = query.filter(idta__gt=idtalastrun)
    if 'frompartner' in cleaned_data and cleaned_data['frompartner']:
        query = frompartnerquery(query,cleaned_data.pop('frompartner'))
    if 'topartner' in cleaned_data and cleaned_data['topartner']:
        query = topartnerquery(query,cleaned_data.pop('topartner'))
    if 'infilename' in cleaned_data and cleaned_data['infilename']:
        query = query.filter(infilename__contains=cleaned_data.pop('infilename'))
    if 'filename' in cleaned_data and cleaned_data['filename']:
        query = query.filter(filename__contains=cleaned_data.pop('filename'))
    for key,value in cleaned_data.items():
        if not value:
            del cleaned_data[key]
    query = query.filter(**cleaned_data)
    if paginate:
        paginator = Paginator(query, botsglobal.ini.getint('settings','limit',30))
        try:
            return paginator.page(page)
        except (EmptyPage, InvalidPage):  #page does not exist: use last page
            lastpage = paginator.num_pages
            org_cleaned_data['page'] = lastpage  #change value in form as well
            return paginator.page(lastpage)
    else:   #do not use paginator; return everything
        return query

def frompartnerquery(query,idpartner):
    # return the appropriate query according to partner type
    # if group: select partners in the group
    # else: select the partner
    isgroup = models.partner.objects.values_list('isgroup', flat=True).filter(idpartner=idpartner)
    if isgroup[0]:
        return query.filter(frompartner__in=models.partner.objects.values_list('idpartner', flat=True).filter(group=idpartner))
    else:
        return query.filter(frompartner=idpartner)

def topartnerquery(query,idpartner):
    isgroup = models.partner.objects.values_list('isgroup', flat=True).filter(idpartner=idpartner)
    if isgroup[0]:
        return query.filter(topartner__in=models.partner.objects.values_list('idpartner', flat=True).filter(group=idpartner))
    else:
        return query.filter(topartner=idpartner)


def indent_x12(content):
    if content.count('\n') > 6:
        return content
    count = 0
    for char in content[:200].lstrip():
        if char in '\r\n' and count != 105: #pos 105: is record_sep, could be \r\n
            continue
        count += 1
        if count == 106:
            sep = char
            break
    else:
        return content
    if sep.isalnum() or sep.isspace():
        return content
    return content.replace(sep,sep + '\n')

EDIFACT_INDENT = re.compile('''
    (?<!\?)     #if no preceding escape (?)
    '           #apostrophe
    (?![\n\r])  #if no following CR of LF
    ''',re.VERBOSE)
def indent_edifact(content):
    ''' three options:
        1 is already indented (each segment on separate line)
        2 block mode: lines of fixed lengths eg 80 chars
        3 one string, one line
    '''
    if content.count('\n') > 4:
        return content
    return EDIFACT_INDENT.sub("'\n", content)



import os
import re
import zipfile
import string
from django.utils.translation import ugettext as _
#bots-modules
import botslib
import botsglobal
from botsconfig import *

@botslib.log_session
def preprocess(routedict,function,status=FILEIN,rootidta=None,**argv):
    ''' for preprocessing of files.
        these are NOT translations; translation involve grammars, mapping scripts etc. think of eg:
        - unzipping zipped files.
        - password protected files.
        Than the actual processing function is called.
        If errors occur during processing, no ta are left with status FILEIN !
        preprocess is called right after the in-communicatiation
    '''
    if rootidta is None:
        rootidta = botsglobal.currentrun.get_minta4query()
    nr_files = 0
    for row in botslib.query(u'''SELECT idta,filename
                                FROM ta
                                WHERE idta>%(rootidta)s
                                AND status=%(status)s
                                AND statust=%(statust)s
                                AND idroute=%(idroute)s
                                AND fromchannel=%(fromchannel)s
                                ''',
                                {'status':status,'statust':OK,'idroute':routedict['idroute'],'fromchannel':routedict['fromchannel'],'rootidta':rootidta}):
        try:
            botsglobal.logger.debug(u'Start preprocessing "%(name)s" for file "%(filename)s".',
                                    {'name':function.__name__,'filename':row['filename']})
            ta_from = botslib.OldTransaction(row['idta'])
            ta_from.filename = row['filename']
            function(ta_from=ta_from,endstatus=status,routedict=routedict,**argv)
        except:
            txt = botslib.txtexc()
            ta_from.update(statust=ERROR,errortext=txt)
            ta_from.deletechildren()
        else:
            botsglobal.logger.debug(u'OK preprocessing "%(name)s" for file "%(filename)s".',
                                    {'name':function.__name__,'filename':row['filename']})
            ta_from.update(statust=DONE)
            nr_files += 1
    return nr_files
    
@botslib.log_session
def postprocess(routedict,function,status=FILEOUT,rootidta=None,**argv):
    ''' for postprocessing of files.
        these are NOT translations; translation involve grammars, mapping scripts etc. think of eg:
        - zip files.
        If errors occur during processing, no ta are left with status FILEOUT !
        postprocess is called right before the out-communicatiation
    '''
    if rootidta is None:
        rootidta = botsglobal.currentrun.get_minta4query_routepart()
    nr_files = 0
    for row in botslib.query(u'''SELECT idta,filename
                                FROM ta
                                WHERE idta>%(rootidta)s
                                AND status=%(status)s
                                AND statust=%(statust)s
                                AND idroute=%(idroute)s
                                AND tochannel=%(tochannel)s
                                ''',
                                {'status':status,'statust':OK,'idroute':routedict['idroute'],'tochannel':routedict['tochannel'],'rootidta':rootidta}):
        try:
            botsglobal.logger.debug(u'Start postprocessing "%(name)s" for file "%(filename)s".',
                                    {'name':function.__name__,'filename':row['filename']})
            ta_from = botslib.OldTransaction(row['idta'])
            ta_from.filename = row['filename']
            function(ta_from=ta_from,endstatus=status,routedict=routedict,**argv)
        except:
            txt = botslib.txtexc()
            ta_from.update(statust=ERROR,errortext=txt)
            ta_from.deletechildren()
        else:
            botsglobal.logger.debug(u'OK postprocessing "%(name)s" for file "%(filename)s".',
                                    {'name':function.__name__,'filename':row['filename']})
            ta_from.update(statust=DONE)
            nr_files += 1
    return nr_files

#regular expression for mailbag.
HEADER = re.compile('''
    \s*
    (
        (?P<edifact>
            (?P<UNA>
                U[\n\r]*N[\n\r]*A
                (?P<UNAstring>.+?)
            )?
            (?P<UNB>
                U[\n\r]*N[\n\r]*B
            )
            [\n\r]*
            (?P<field_sep>[^\n\r])
        )
        |
        (?P<tradacoms>
            (?P<STX>
                S[\n\r]*T[\n\r]*X
                [\n\r]*
                =
            )
        )
        |
        (?P<x12>
            I[\n\r]*S[\n\r]*A
        )
    )
    ''',re.DOTALL|re.VERBOSE)

def mailbag(ta_from,endstatus,frommessagetype,**argv):
    ''' 2 main functions:
        -   recognizes and distuinguishes several edi types: x12 edifact tradacoms ('mailbag' in, correct editype out)
        -   split up interchanges (edifact, x12, tradacoms)
        details:
        - edifact, x12 and tradacoms can be can be mixed
        - handle multiple UNA in one file, including different charsets.
        - handle multiple ISA's with different separators in one file
        in bots > 3.0.0 all mailbag, edifact, x12 and tradacoms go via mailbag.
    '''
    edifile = botslib.readdata(filename=ta_from.filename)
    startpos = 0
    nr_interchanges = 0
    while (1):
        found = HEADER.match(edifile[startpos:])
        if found is None:
            if edifile[startpos:].strip(string.whitespace+'\x1A\x00'):  #there is content...but not valid
                if nr_interchanges:    #found interchanges, but remainder is not valid
                    raise botslib.InMessageError(_(u'[M50]: Found data not in a valid interchange at position %(pos)s.'),{'pos':startpos})                
                else:   #no interchanges found, content is not a valid edifact/x12/tradacoms interchange
                    if frommessagetype == 'mailbag':    #if indicated 'mailbag': guess if this is an xml file.....
                        sniffxml = edifile[:25]
                        sniffxml = sniffxml.lstrip(' \t\n\r\f\v\xFF\xFE\xEF\xBB\xBF\x00')       #to find first ' real' data; some char are because of BOM, UTF-16 etc
                        if sniffxml and sniffxml[0] == '<':
                            #is a xml file; inmessage.py can determine the right xml messagetype via xpath. 
                            filesize = len(edifile)
                            ta_to = ta_from.copyta(status=endstatus,statust=OK,filename=ta_from.filename,editype='xml',messagetype='mailbag',filesize=filesize)
                            return
                    raise botslib.InMessageError(_(u'[M51]: Edi file does not start with a valid interchange.'))
            else:   #no parseble content
                if nr_interchanges:    #OK: there are interchanges, but no new interchange is found.
                    return
                else:   #no edifact/x12/tradacoms envelope at all
                    raise botslib.InMessageError(_(u'[M52]: Edi file contains only whitespace.'))
        elif found.group('x12'):
            editype = 'x12'
            headpos = startpos + found.start('x12')
            #determine field_sep and record_sep
            count = 0
            for char in edifile[headpos:headpos+120]:  #search first 120 characters to determine separators
                if char in '\r\n' and count != 105:
                    continue
                count += 1
                if count == 4:
                    field_sep = char
                elif count in [7,18,21,32,35,51,54,70]:   #extra checks for fixed ISA. 
                    if char != field_sep:
                        raise botslib.InMessageError(_(u'[M53]: Non-valid ISA header at position %(pos)s; position %(pos_element)s of ISA is "%(foundchar)s", expect here element separator "%(field_sep)s".'),
                                                        {'pos':headpos,'pos_element':unicode(count),'foundchar':char,'field_sep':field_sep})
                elif count == 106:
                    record_sep = char
                    break
            foundtrailer = re.search('''%(record_sep)s
                                        \s*
                                        I[\n\r]*E[\n\r]*A
                                        .+?
                                        %(record_sep)s
                                        '''%{'record_sep':re.escape(record_sep)},
                                        edifile[headpos:],re.DOTALL|re.VERBOSE)
            if not foundtrailer:
                foundtrailer2 = re.search('''%(record_sep)s
                                            \s*
                                            I[\n\r]*E[\n\r]*A
                                            '''%{'record_sep':re.escape(record_sep)},
                                            edifile[headpos:],re.DOTALL|re.VERBOSE)
                if foundtrailer2:
                    raise botslib.InMessageError(_(u'[M60]: Found no segment terminator for IEA trailer at position %(pos)s.'),{'pos':foundtrailer2.start()})
                else:
                    raise botslib.InMessageError(_(u'[M54]: Found no valid IEA trailer for the ISA header at position %(pos)s.'),{'pos':headpos})
        elif found.group('edifact'):
            editype = 'edifact'
            headpos = startpos + found.start('edifact')
            #parse UNA. valid UNA: UNA:+.? '
            if found.group('UNA'):
                count = 0
                for char in found.group('UNAstring'):
                    if char in '\r\n':
                        continue
                    count += 1
                    if count == 2:
                        field_sep = char
                    elif count == 4:
                        escape = char
                    elif count == 6:
                        record_sep = char
                if count != 6 and len(found.group('UNAstring').rstrip()) != 6:
                    raise botslib.InMessageError(_(u'[M55]: Non-valid UNA-segment at position %(pos)s. UNA-segment should be 6 positions.'),{'pos':headpos})
                if found.group('field_sep') != field_sep:
                    raise botslib.InMessageError(_(u'[M56]: Data element separator used in edifact file differs from value indicated in UNA-segment.'))
            else:   #no UNA, interpret UNB
                if found.group('field_sep') == '+':
                    record_sep = "'"
                    escape = '?'
                elif found.group('field_sep') == '\x1D':        #according to std this was preffered way...probably quite theoretic...but does no harm
                    record_sep = '\x1C'
                    escape = ''
                else:
                    raise botslib.InMessageError(_(u'[M57]: Edifact file with non-standard separators. UNA segment should be used.'))
            #search trailer
            foundtrailer = re.search('''[^%(escape)s\n\r]       #char that is not escape or cr/lf
                                        [\n\r]*?                #maybe some cr/lf's
                                        %(record_sep)s          #segment separator
                                        \s*                     #whitespace between segments
                                        U[\n\r]*N[\n\r]*Z       #UNZ
                                        .+?                     #any chars
                                        [^%(escape)s\n\r]       #char that is not escape or cr/lf
                                        [\n\r]*?                #maybe some cr/lf's
                                        %(record_sep)s          #segment separator
                                        '''%{'escape':escape,'record_sep':re.escape(record_sep)},
                                        edifile[headpos:],re.DOTALL|re.VERBOSE)
            if not foundtrailer:
                raise botslib.InMessageError(_(u'[M58]: Found no valid UNZ trailer for the UNB header at position %(pos)s.'),{'pos':headpos})
        elif found.group('tradacoms'):
            editype = 'tradacoms'
            #~ field_sep = '='     #the tradacoms 'after-segment-tag-separator'
            record_sep = "'"
            escape = '?'
            headpos = startpos + found.start('STX')
            foundtrailer = re.search('''[^%(escape)s\n\r]       #char that is not escape or cr/lf
                                        [\n\r]*?                #maybe some cr/lf's
                                        %(record_sep)s          #segment separator
                                        \s*                     #whitespace between segments
                                        E[\n\r]*N[\n\r]*D
                                        .+?
                                        [^%(escape)s\n\r]       #char that is not escape or cr/lf
                                        [\n\r]*?                #maybe some cr/lf's
                                        %(record_sep)s          #segment separator
                                        '''%{'escape':escape,'record_sep':re.escape(record_sep)},
                                        edifile[headpos:],re.DOTALL|re.VERBOSE)
            if not foundtrailer:
                raise botslib.InMessageError(_(u'[M59]: Found no valid END trailer for the STX header at position %(pos)s.'),{'pos':headpos})
        #so: found an interchange (from headerpos until endpos)
        endpos = headpos + foundtrailer.end()
        ta_to = ta_from.copyta(status=endstatus)  #make transaction for translated message; gets ta_info of ta_frommes
        tofilename = unicode(ta_to.idta)
        filesize = len(edifile[headpos:endpos])
        tofile = botslib.opendata(tofilename,'wb')
        tofile.write(edifile[headpos:endpos])
        tofile.close()
        #editype is now either edifact, x12 or tradacoms
        #frommessagetype is the original frommessagetype (from route).
        #frommessagetype would normally be edifact, x12, tradacoms or mailbag, but could also be eg ORDERSD96AUNEAN007.
        #If so, we want to preserve that.
        if frommessagetype != 'mailbag' and frommessagetype != editype:
            messagetype = frommessagetype
        else:
            messagetype = editype
        ta_to.update(statust=OK,filename=tofilename,editype=editype,messagetype=messagetype,filesize=filesize) #update outmessage transaction with ta_info;
        startpos = endpos
        nr_interchanges += 1
        botsglobal.logger.debug(_(u'        File written: "%(tofilename)s".'),{'tofilename':tofilename})


def botsunzip(ta_from,endstatus,password=None,pass_non_zip=False,**argv):
    ''' unzip file;
        editype & messagetype are unchanged.
    '''
    try:
        myzipfile = zipfile.ZipFile(botslib.abspathdata(filename=ta_from.filename),mode='r')
    except zipfile.BadZipfile:
        botsglobal.logger.debug(_(u'File is not a zip-file.'))
        if pass_non_zip:        #just pass the file
            botsglobal.logger.debug(_(u'"pass_non_zip" is True, just pass the file.'))
            ta_to = ta_from.copyta(status=endstatus,statust=OK)
            return
        raise botslib.InMessageError(_(u'File is not a zip-file.'))

    if password:
        myzipfile.setpassword(password)
    for info_file_in_zip in myzipfile.infolist():
        if info_file_in_zip.filename[-1] == '/':    #check if this is a dir; if so continue
            continue
        ta_to = ta_from.copyta(status=endstatus)
        tofilename = unicode(ta_to.idta)
        content = myzipfile.read(info_file_in_zip.filename)    #read file in zipfile
        filesize = len(content)
        tofile = botslib.opendata(tofilename,'wb')
        tofile.write(content)
        tofile.close()
        ta_to.update(statust=OK,filename=tofilename,filesize=filesize) #update outmessage transaction with ta_info;
        botsglobal.logger.debug(_(u'        File written: "%(tofilename)s".'),{'tofilename':tofilename})
    myzipfile.close()

def botszip(ta_from,endstatus,**argv):
    ''' zip file;
        editype & messagetype are unchanged.
    '''
    ta_to = ta_from.copyta(status=endstatus)
    tofilename = unicode(ta_to.idta)
    pluginzipfilehandler = zipfile.ZipFile(botslib.abspathdata(filename=tofilename), 'w', zipfile.ZIP_DEFLATED)
    pluginzipfilehandler.write(botslib.abspathdata(filename=ta_from.filename),ta_from.filename)
    pluginzipfilehandler.close()
    ta_to.update(statust=OK,filename=tofilename) #update outmessage transaction with ta_info;


def extractpdf(ta_from,endstatus,**argv):
    ''' Try to extract text content of a PDF file to a csv.
        You know this is not a great idea, right? But we'll do the best we can anyway!
        Page and line numbers are added to each row.
        Columns and rows are based on the x and y coordinates of each text element within tolerance allowed.
        Multiple text elements may combine to make one field, some PDFs have every character separated!
        You may need to experiment with x_group and y_group values, but defaults seem ok for most files.
        Output csv is UTF-8 encoded - The csv module doesn't directly support reading and writing Unicode
        If the PDF is just an image, all bets are off. Maybe try OCR, good luck with that!
        Mike Griffin 14/12/2011
    '''
    from pdfminer.pdfinterp import PDFResourceManager, process_pdf
    from pdfminer.converter import TextConverter
    from pdfminer.layout import LAParams, LTContainer, LTText, LTTextBox
    import csv

    class CsvConverter(TextConverter):
        def __init__(self, *args, **kwargs):
            TextConverter.__init__(self, *args, **kwargs)

        def receive_layout(self, ltpage):

            # recursively get every text element and it's coordinates
            def render(item):
                if isinstance(item, LTContainer):
                    for child in item:
                        render(child)
                elif isinstance(item, LTText):
                    (unused1,unused2,x,y) = item.bbox

                    # group the y values (rows) within group tolerance
                    for v in yv:
                        if y > v-y_group and y < v+y_group:
                            y = v
                    yv.append(y)

                    line = lines[int(-y)]
                    line[x] = item.get_text().encode('utf-8')

            from collections import defaultdict
            lines = defaultdict(lambda : {})

            yv = []
            render(ltpage)

            lineid = 0
            for y in sorted(lines.keys()):
                line = lines[y]
                lineid += 1
                csvdata = [ltpage.pageid,lineid] # first 2 columns are page and line numbers

                # group the x values (fields) within group tolerance
                p = 0
                field_txt = ''
                for x in sorted(line.keys()):
                    gap = x - p
                    if p > 0 and gap > x_group:
                        csvdata.append(field_txt)
                        field_txt = ''
                    field_txt += line[x]
                    p = x
                csvdata.append(field_txt)
                csvout.writerow(csvdata)
            if lineid == 0:
                raise botslib.InMessageError(_(u'PDF text extraction failed, it may contain just image(s)?'))


    #get some optional parameters
    x_group = argv.get('x_group',10) # group text closer than this as one field
    y_group = argv.get('y_group',5)  # group lines closer than this as one line
    password = argv.get('password','')
    quotechar = argv.get('quotechar','"')
    field_sep = argv.get('field_sep',',')
    escape = argv.get('escape','\\')
    charset = argv.get('charset','utf-8')
    if not escape:
        doublequote = True
    else:
        doublequote = False

    try:
        pdf_stream = botslib.opendata(ta_from.filename, 'rb')
        ta_to = ta_from.copyta(status=endstatus)
        tofilename = unicode(ta_to.idta)
        csv_stream = botslib.opendata(tofilename,'wb')
        csvout = csv.writer(csv_stream, quotechar=quotechar, delimiter=field_sep, doublequote=doublequote, escapechar=escape)

        # Process PDF
        rsrcmgr = PDFResourceManager(caching=True)
        device = CsvConverter(rsrcmgr, csv_stream, codec=charset)
        process_pdf(rsrcmgr, device, pdf_stream, pagenos=set(), password=password, caching=True, check_extractable=True)

        device.close()
        pdf_stream.close()
        csv_stream.close()
        filesize = os.path.getsize(botslib.abspathdata(tofilename))
        ta_to.update(statust=OK,filename=tofilename,filesize=filesize) #update outmessage transaction with ta_info;
        botsglobal.logger.debug(_(u'        File written: "%(tofilename)s".'),{'tofilename':tofilename})
    except:
        txt = botslib.txtexc()
        botsglobal.logger.error(_(u'PDF extraction failed, may not be a PDF file? Error:\n%(txt)s'),{'txt':txt})
        raise botslib.InMessageError(_(u'PDF extraction failed, may not be a PDF file? Error:\n%(txt)s'),{'txt':txt})





#!/usr/bin/env python
import sys
import os
import django
from django.core.handlers.wsgi import WSGIHandler
from django.utils.translation import ugettext as _
import cherrypy
from cherrypy import wsgiserver
import botsglobal
import botsinit


def start():
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #***command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    The %(name)s is the web server for bots; the interface (bots-monitor) can be accessed in a 
    browser, eg 'http://localhost:8080'.
    Usage:
        %(name)s  -c<directory>
    Options:
        -c<directory>   directory for configuration files (default: config).
    
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        else:
            print usage
            sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    process_name = 'webserver'
    botsglobal.logger = botsinit.initserverlogging(process_name)    #initialise file-logging for web-server. This logging only contains the logging from bots-webserver, not from cherrypy.

    #***init cherrypy as webserver*********************************************
    #global configuration for cherrypy
    cherrypy.config.update({'global': {'log.screen': False, 'server.environment': botsglobal.ini.get('webserver','environment','production')}})
    #cherrypy handling of static files
    conf = {'/': {'tools.staticdir.on' : True,'tools.staticdir.dir' : 'media' ,'tools.staticdir.root': botsglobal.ini.get('directories','botspath')}}
    servestaticfiles = cherrypy.tree.mount(None, '/media', conf)    #None: no cherrypy application (as this only serves static files)
    #cherrypy handling of django
    servedjango = WSGIHandler()     #was: servedjango = AdminMediaHandler(WSGIHandler())  but django does not need the AdminMediaHandler in this setup. is much faster.
    #cherrypy uses a dispatcher in order to handle the serving of static files and django.
    dispatcher = wsgiserver.WSGIPathInfoDispatcher({'/': servedjango, '/media': servestaticfiles})
    botswebserver = wsgiserver.CherryPyWSGIServer(bind_addr=('0.0.0.0', botsglobal.ini.getint('webserver','port',8080)), wsgi_app=dispatcher, server_name=botsglobal.ini.get('webserver','name','bots-webserver'))
    botsglobal.logger.log(25,_(u'Bots %(process_name)s started.'),
                                {'process_name':process_name})
    botsglobal.logger.log(25,_(u'Bots %(process_name)s configdir: "%(configdir)s".'),
                                {'process_name':process_name, 'configdir':botsglobal.ini.get('directories','config')})
    botsglobal.logger.log(25,_(u'Bots %(process_name)s serving at port: "%(port)s".'),
                                {'process_name':process_name,'port':botsglobal.ini.getint('webserver','port',8080)})
    #handle ssl: cherrypy < 3.2 always uses pyOpenssl. cherrypy >= 3.2 uses python buildin ssl (python >= 2.6 has buildin support for ssl).
    ssl_certificate = botsglobal.ini.get('webserver','ssl_certificate',None)
    ssl_private_key = botsglobal.ini.get('webserver','ssl_private_key',None)
    if ssl_certificate and ssl_private_key:
        if cherrypy.__version__ >= '3.2.0':
            adapter_class = wsgiserver.get_ssl_adapter_class('builtin')
            botswebserver.ssl_adapter = adapter_class(ssl_certificate,ssl_private_key)
        else:
            #but: pyOpenssl should be there!
            botswebserver.ssl_certificate = ssl_certificate
            botswebserver.ssl_private_key = ssl_private_key
        botsglobal.logger.log(25,_(u'Bots %(process_name)s uses ssl (https).'),{'process_name':process_name})
    else:
        botsglobal.logger.log(25,_(u'Bots %(process_name)s uses plain http (no ssl).'),{'process_name':process_name})

    #***start the cherrypy webserver.************************************************
    try:
        botswebserver.start()
    except KeyboardInterrupt:
        botswebserver.stop()


if __name__ == '__main__':
    start()



''' converts xml file to a bots grammar.
    Usage: c:\python27\python  bots-xml2botsgrammar.py  botssys/infile/test.xml   botssys/infile/resultgrammar.py  -cconfig
    Try to have a 'completely filled' xml file.
'''
import os
import sys
import atexit
import copy
import logging
try:
    import cElementTree as ET
except ImportError:
    try:
        import elementtree.ElementTree as ET
    except ImportError:
        try:
            from xml.etree import cElementTree as ET
        except ImportError:
            from xml.etree import ElementTree as ET
try:
    from collections import OrderedDict
except:
    from bots_ordereddict import OrderedDict    
import botslib
import botsinit
import botsglobal
import inmessage
import outmessage
import node
from botsconfig import *

#**************************************************************************************
#***classes used in inmessage for xml2botsgrammar.
#***These classes are dynamically added to inmessage
#**************************************************************************************
class xmlforgrammar(inmessage.Inmessage):
    ''' class for ediobjects in XML. Uses ElementTree'''
    def initfromfile(self):
        filename = botslib.abspathdata(self.ta_info['filename'])
        self.ta_info['attributemarker'] = '__'
        parser = ET.XMLParser()
        etree =  ET.ElementTree()   #ElementTree: lexes, parses, makes etree; etree is quite similar to bots-node trees but conversion is needed
        etreeroot = etree.parse(filename, parser)
        self.root = self._etree2botstree(etreeroot)  #convert etree to bots-nodes-tree
            
    def _use_botscontent(self,xmlnode):
        if self._is_record(xmlnode):
            if xmlnode.text is None:
                return False
            else:
                return not xmlnode.text.isspace()
        else:
            return True
            
    def _etree2botstree(self,xmlnode):
        newnode = node.Node(record=self._etreenode2botstreenode(xmlnode))
        for xmlchildnode in xmlnode:   #for every node in mpathtree
            if self._is_record(xmlchildnode):
                newnode.append(self._etree2botstree(xmlchildnode))           #add as a node/record
            else: 
                ## remark for generating grammars: empty strings should generate a field here
                if self._use_botscontent(xmlchildnode):
                    newnode.record[xmlchildnode.tag] = '1'      #add as a field
                #convert the xml-attributes of this 'xml-field' to fields in dict with attributemarker.
                newnode.record.update((xmlchildnode.tag + self.ta_info['attributemarker'] + key, value) for key,value in xmlchildnode.items())
        return newnode

    def _etreenode2botstreenode(self,xmlnode):
        ''' build a OrderedDict from xml-node. Add BOTSID, xml-attributes (of 'record'), xmlnode.text as BOTSCONTENT.'''
        build = OrderedDict((xmlnode.tag + self.ta_info['attributemarker'] + key,value) for key,value in xmlnode.items())   #convert xml attributes to fields.
        build['BOTSID'] = xmlnode.tag
        if self._use_botscontent(xmlnode):
            build['BOTSCONTENT'] = '1'
        return build

    def _is_record(self,xmlchildnode):
        return bool(len(xmlchildnode))


class xmlforgrammar_allrecords(inmessage.Inmessage):
    ''' class for ediobjects in XML. Uses ElementTree'''
    def initfromfile(self):
        filename = botslib.abspathdata(self.ta_info['filename'])
        self.ta_info['attributemarker'] = '__'
        parser = ET.XMLParser()
        etree =  ET.ElementTree()   #ElementTree: lexes, parses, makes etree; etree is quite similar to bots-node trees but conversion is needed
        etreeroot = etree.parse(filename, parser)
        self.root = self._etree2botstree(etreeroot)  #convert etree to bots-nodes-tree
            
    def _etree2botstree(self,xmlnode):
        newnode = node.Node(record=self._etreenode2botstreenode(xmlnode))
        for xmlchildnode in xmlnode:   #for every node in mpathtree
            newnode.append(self._etree2botstree(xmlchildnode))           #add as a node/record
        return newnode

    def _etreenode2botstreenode(self,xmlnode):
        ''' build a OrderedDict from xml-node. Add BOTSID, xml-attributes (of 'record'), xmlnode.text as BOTSCONTENT.'''
        build = OrderedDict((xmlnode.tag + self.ta_info['attributemarker'] + key,value) for key,value in xmlnode.items())   #convert xml attributes to fields.
        build['BOTSID'] = xmlnode.tag
        if not self._is_record(xmlnode):
            build['BOTSCONTENT'] = '1'
        return build

    def _is_record(self,xmlchildnode):
        return bool(len(xmlchildnode))

#******************************************************************
#***functions for mapping******************************************
def map_treewalker(node_instance,mpath):
    ''' Generator function.
    '''
    mpath.append(OrderedDict({'BOTSID':node_instance.record['BOTSID']}))
    for childnode in node_instance.children:
        yield childnode,mpath[:]
        for terugnode,terugmpath in map_treewalker(childnode,mpath):
            yield terugnode,terugmpath
    mpath.pop()


def map_writefields(node_out,node_in,mpath):
    ''' als fields of this level are written to node_out.
    '''
    mpath_with_all_fields = copy.deepcopy(mpath)     #use a copy of mpath (do not want to change it)
    for key in node_in.record.keys():       
        if key in ['BOTSID','BOTSIDnr']:    #skip these
            continue
        mpath_with_all_fields[-1][key] = u'dummy'    #add key to the mpath
    node_out.put(*mpath_with_all_fields)            #write all fields.

#******************************************************************
#***functions to convert out-tree to grammar*********************
def tree2grammar(node_instance,structure,recorddefs):
    structure.append({ID:node_instance.record['BOTSID'],MIN:0,MAX:99999,LEVEL:[]})
    recordlist = []
    for key in node_instance.record.keys():
        recordlist.append([key, 'C', 256, 'AN'])
    if node_instance.record['BOTSID'] in recorddefs:
        recorddefs[node_instance.record['BOTSID']] = removedoublesfromlist(recorddefs[node_instance.record['BOTSID']] + recordlist)
    else:
        recorddefs[node_instance.record['BOTSID']] = recordlist
    for childnode in node_instance.children:
        tree2grammar(childnode,structure[-1][LEVEL],recorddefs)

def removedoublesfromlist(orglist):
    list2return = []
    for member in orglist:
        if member not in list2return:
            list2return.append(member)
    return list2return

#******************************************************************
#***functions to write grammar to file*****************************
def recorddefs2string(recorddefs,targetNamespace):
    result = ""
    for tag in sorted(recorddefs):
        result += "'%s%s':\n    [\n"%(targetNamespace,tag)
        for field in recorddefs[tag]:
            if field[0] in ['BOTSID','BOTSCONTENT']:
                field[1] = 'M'
                result +=  "    ['%s', '%s', %s, '%s'],\n"%(field[0],field[1],field[2],field[3])
        for field in recorddefs[tag]:
            if field[0].startswith(tag + '__'):
                result +=  "    ['%s', '%s', %s, '%s'],\n"%(field[0],field[1],field[2],field[3])
        for field in recorddefs[tag]:
            if field[0] not in ['BOTSID','BOTSIDnr','BOTSCONTENT'] and not field[0].startswith(tag + '__'):
                result += "    ['%s%s', '%s', %s, '%s'],\n"%(targetNamespace,field[0],field[1],field[2],field[3])
        result += "    ],\n"
    return result

def structure2string(structure,targetNamespace,level=0):
    result = ""
    for segment in structure:
        if LEVEL in segment and segment[LEVEL]:
            result += level*'    ' + "{ID:'%s%s',MIN:%s,MAX:%s,LEVEL:[\n"%(targetNamespace,segment[ID],segment[MIN],segment[MAX])
            result += structure2string(segment[LEVEL],targetNamespace,level+1)
            result += level*'    ' + "]},\n"
        else:
            result += level*'    ' + "{ID:'%s%s',MIN:%s,MAX:%s},\n"%(targetNamespace,segment[ID],segment[MIN],segment[MAX])
    return result

def grammar2file(botsgrammarfilename,structure,recorddefs,targetNamespace):
    if targetNamespace:
        targetNamespace = '{' + targetNamespace + '}'
    result = 'structure = [\n'
    result += structure2string(structure,targetNamespace)
    result += ']\n\n'
    result += 'recorddefs = {\n'
    result += recorddefs2string(recorddefs,targetNamespace)
    result +=  "}\n"
    
    result2 = '#Generated by bots open source edi translator.\nfrom bots.botsconfig import *\n'
    if targetNamespace:
        shortNamespace = 'NS'
        result2 += shortNamespace + " = '" + targetNamespace + "'\n\n"
        result = result.replace("'" + targetNamespace,shortNamespace + "+'")
    result2 += result
    
    f = open(botsgrammarfilename,'wb')
    f.write(result2)
    f.close()
    print 'grammar file is written:',botsgrammarfilename


def start():
    #********command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    Creates a grammar from an xml file.'
    Usage:'
        %(name)s  -c<directory>  <xml_file>  <xml_grammar_file>
    Options:
        -c<directory>      directory for configuration files (default: config).
        -a                 all xml elements as records
        <xml_file>         name of the xml file to read
        <xml_grammar_file> name of the grammar file to write
    
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    edifile =''
    botsgrammarfilename = ''
    allrecords = False
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        elif arg.startswith('-a'):
            allrecords = True
        elif arg in ["?", "/?",'-h', '--help'] or arg.startswith('-'):
            print usage
            sys.exit(0)
        else:
            if not edifile:
                edifile = arg
            else:
                botsgrammarfilename = arg
    if not edifile or not botsgrammarfilename:
        print 'Error: both edifile and grammarfile are required.'
        sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    process_name = 'xml2botsgrammar'
    botsglobal.logger = botsinit.initenginelogging(process_name)
    atexit.register(logging.shutdown)

    targetNamespace = ''
    #*******************************************************************
    #***add classes for handling editype xml to inmessage
    #*******************************************************************
    if allrecords:
        #~ editype = 'xmlforgrammar_allrecords'
        inmessage.xmlforgrammar = xmlforgrammar_allrecords
    else:
        #~ editype = 'xmlforgrammar' 
        inmessage.xmlforgrammar = xmlforgrammar
    #make inmessage object: read the xml file
    inn = inmessage.parse_edi_file(editype='xmlforgrammar',messagetype='',filename=edifile)
    inn.checkforerrorlist() #no exception if infile has been lexed and parsed OK else raises an error
    #make outmessage object; nothing is 'filled' yet. In mapping tree is filled; nothing is written to file.
    out = outmessage.outmessage_init(editype='xmlnocheck',messagetype='',filename='',divtext='',topartner='')    
    
    #***mapping: make 'normalised' out-tree suited for writing as a grammar***************************************************
    mpath_root = [OrderedDict({'BOTSID':inn.root.record['BOTSID'],'BOTSIDnr':'1'})] #handle root
    out.put(*mpath_root)
    map_writefields(out,inn.root,mpath_root)
    
    #walk tree; write results to out-tree
    mpath_start = []
    for node_instance,mpath in map_treewalker(inn.root,mpath_start):
        mpath.append(OrderedDict({'BOTSID':node_instance.record['BOTSID']}))
        if out.get(*mpath) is None:     #if node does not exist: write it.
            out.put(*mpath)
        map_writefields(out,node_instance,mpath)
    #***mapping is done 

    #***convert out-tree to grammar
    structure = []
    recorddefs = {}
    tree2grammar(out.root,structure,recorddefs)

    #***write grammar to file
    grammar2file(botsgrammarfilename,structure,recorddefs,targetNamespace)
    

if __name__ == '__main__':
    start()



''' functions for reading and making plugins.
    Reading an making functions are seperate functions.
'''
import os
import sys
#~ import time
import zipfile
import zipimport
import codecs
import django
from django.core import serializers
from django.utils.translation import ugettext as _
import models
import botslib
import botsglobal

#******************************************
#* read a plugin **************************
#******************************************
@django.db.transaction.commit_on_success  #if no exception raised: commit, else rollback.
def read_index(filename):
    ''' process index file in default location. '''
    try:
        importedbotsindex,scriptname = botslib.botsimport('index')
        pluglist = importedbotsindex.plugins[:]
        if 'botsindex' in sys.modules:
            del sys.modules['botsindex']
    except:
        txt = botslib.txtexc()
        raise botslib.PluginError(_(u'Error in configuration index file. Nothing is written. Error:\n%(txt)s'),{'txt':txt})
    else:
        botsglobal.logger.info(_(u'Configuration index file is OK.'))
        botsglobal.logger.info(_(u'Start writing to database.'))

    #write content of index file to the bots database
    try:
        read_index2database(pluglist)
    except:
        txt = botslib.txtexc()
        raise botslib.PluginError(_(u'Error writing configuration index to database. Nothing is written. Error:\n%(txt)s'),{'txt':txt})
    else:
        botsglobal.logger.info(_(u'Writing to database is OK.'))


@django.db.transaction.commit_on_success  #if no exception raised: commit, else rollback.
def read_plugin(pathzipfile):
    ''' process uploaded plugin. '''
    #test if valid zipfile
    if not zipfile.is_zipfile(pathzipfile):
        raise botslib.PluginError(_(u'Plugin is not a valid file.'))

    #read index file
    try:
        myzipimport = zipimport.zipimporter(pathzipfile)
        importedbotsindex = myzipimport.load_module('botsindex')
        pluglist = importedbotsindex.plugins[:]
        if 'botsindex' in sys.modules:
            del sys.modules['botsindex']
    except:
        txt = botslib.txtexc()
        raise botslib.PluginError(_(u'Error in plugin. Nothing is written. Error:\n%(txt)s'),{'txt':txt})
    else:
        botsglobal.logger.info(_(u'Plugin is OK.'))
        botsglobal.logger.info(_(u'Start writing to database.'))

    #write content of index file to the bots database
    try:
        read_index2database(pluglist)
    except:
        txt = botslib.txtexc()
        raise botslib.PluginError(_(u'Error writing plugin to database. Nothing is written. Error:\n%(txt)s'),{'txt':txt})
    else:
        botsglobal.logger.info(_(u'Writing to database is OK.'))

    #write files to the file system.
    botsglobal.logger.info(_(u'Start writing to files'))
    try:
        warnrenamed = False     #to report in GUI files have been overwritten.
        myzip = zipfile.ZipFile(pathzipfile, mode="r")
        orgtargetpath = botsglobal.ini.get('directories','botspath')
        if (orgtargetpath[-1:] in (os.path.sep, os.path.altsep) and len(os.path.splitdrive(orgtargetpath)[1]) > 1):
            orgtargetpath = orgtargetpath[:-1]
        for zipfileobject in myzip.infolist():
            if zipfileobject.filename not in ['botsindex.py','README','botssys/sqlitedb/botsdb','config/bots.ini'] and os.path.splitext(zipfileobject.filename)[1] not in ['.pyo','.pyc']:
                #~ botsglobal.logger.info(u'Filename in zip "%s".',zipfileobject.filename)
                if zipfileobject.filename[0] == '/':
                    targetpath = zipfileobject.filename[1:]
                else:
                    targetpath = zipfileobject.filename
                #convert for correct environment: repacle botssys, config, usersys in filenames
                if targetpath.startswith('usersys'):
                    targetpath = targetpath.replace('usersys',botsglobal.ini.get('directories','usersysabs'),1)
                elif targetpath.startswith('botssys'):
                    targetpath = targetpath.replace('botssys',botsglobal.ini.get('directories','botssys'),1)
                elif targetpath.startswith('config'):
                    targetpath = targetpath.replace('config',botsglobal.ini.get('directories','config'),1)
                targetpath = botslib.join(orgtargetpath, targetpath)
                #targetpath is OK now.
                botsglobal.logger.info(_(u'    Start writing file: "%(targetpath)s".'),{'targetpath':targetpath})

                if botslib.dirshouldbethere(os.path.dirname(targetpath)):
                    botsglobal.logger.info(_(u'        Create directory "%(directory)s".'),{'directory':os.path.dirname(targetpath)})
                if zipfileobject.filename[-1] == '/':    #check if this is a dir; if so continue
                    continue
                if os.path.isfile(targetpath):  #check if file already exists
                    try:    #this ***sometimes*** fails. (python25, for static/help/home.html...only there...)
                        warnrenamed = True
                    except:
                        pass
                source = myzip.read(zipfileobject.filename)
                target = open(targetpath, "wb")
                target.write(source)
                target.close()
                botsglobal.logger.info(_(u'        File written: "%(targetpath)s".'),{'targetpath':targetpath})
    except:
        txt = botslib.txtexc()
        myzip.close()
        raise botslib.PluginError(_(u'Error writing files to system. Nothing is written to database. Error:\n%(txt)s'),{'txt':txt})
    else:
        myzip.close()
        botsglobal.logger.info(_(u'Writing files to filesystem is OK.'))
        return warnrenamed

#PLUGINCOMPARELIST: for filtering and sorting the plugins.
PLUGINCOMPARELIST = ['uniek','persist','mutex','ta','filereport','report','ccodetrigger','ccode', 'channel','partner','chanpar','translate','routes','confirmrule']

def read_index2database(orgpluglist):
    #sanity checks on pluglist
    if not orgpluglist:  #list of plugins is empty: is OK. DO nothing
        return
    if not isinstance(orgpluglist,list):   #has to be a list!!
        raise botslib.PluginError(_(u'Plugins should be list of dicts. Nothing is written.'))
    for plug in orgpluglist:
        if not isinstance(plug,dict):
            raise botslib.PluginError(_(u'Plugins should be list of dicts. Nothing is written.'))
        for key in plug.keys():
            if not isinstance(key,basestring):
                raise botslib.PluginError(_(u'Key of dict is not a string: "%(plug)s". Nothing is written.'),{'plug':plug})
        if 'plugintype' not in plug:
            raise botslib.PluginError(_(u'"Plugintype" missing in: "%(plug)s". Nothing is written.'),{'plug':plug})

    #special case: compatibility with bots 1.* plugins.
    #in bots 1.*, partnergroup was in separate tabel; in bots 2.* partnergroup is in partner
    #later on, partnergroup will get filtered
    for plug in orgpluglist[:]:
        if plug['plugintype'] == 'partnergroup':
            for plugpartner in orgpluglist:
                if plugpartner['plugintype'] == 'partner' and plugpartner['idpartner'] == plug['idpartner']:
                    if 'group' in plugpartner:
                        plugpartner['group'].append(plug['idpartnergroup'])
                    else:
                        plugpartner['group'] = [plug['idpartnergroup']]
                    break

    #copy & filter orgpluglist; do plugtype specific adaptions
    pluglist = []
    for plug in orgpluglist:
        if plug['plugintype'] == 'ccode':   #add ccodetrigger. #20101223: this is NOT needed; codetrigger shoudl be in plugin.
            for seachccodetriggerplug in pluglist:
                if seachccodetriggerplug['plugintype'] == 'ccodetrigger' and seachccodetriggerplug['ccodeid'] == plug['ccodeid']:
                    break
            else:
                pluglist.append({'plugintype':'ccodetrigger','ccodeid':plug['ccodeid']})
        elif plug['plugintype'] == 'translate': #make some fields None instead of '' (translate formpartner, topartner)
            if not plug['frompartner']:
                plug['frompartner'] = None
            if not plug['topartner']:
                plug['topartner'] = None
        elif plug['plugintype'] == 'routes':
            plug['active'] = False
            if 'defer' not in plug:
                plug['defer'] = False
            else:
                if plug['defer'] is None:
                    plug['defer'] = False
        elif plug['plugintype'] == 'channel':
            #convert for correct environment: path and mpath in channels
            if 'path' in plug and plug['path'].startswith('botssys'):
                plug['path'] = plug['path'].replace('botssys',botsglobal.ini.get('directories','botssys_org'),1)
            if 'testpath' in plug and plug['testpath'].startswith('botssys'):
                plug['testpath'] = plug['testpath'].replace('botssys',botsglobal.ini.get('directories','botssys_org'),1)
        elif plug['plugintype'] == 'confirmrule':
            plug.pop('id', None)       #id is an artificial key, delete,
        elif plug['plugintype'] not in PLUGINCOMPARELIST:   #if not in PLUGINCOMPARELIST: do not use
            continue
        pluglist.append(plug)
    #sort pluglist: this is needed for relationships
    pluglist.sort(key=lambda plug: plug.get('isgroup',False),reverse=True)       #sort partners on being partnergroup or not
    pluglist.sort(key=lambda plug: PLUGINCOMPARELIST.index(plug['plugintype']))   #sort all plugs on plugintype; are partners/partenrgroups are already sorted, this will still be true in this new sort (python guarantees!)

    for plug in pluglist:
        botsglobal.logger.info(u'    Start write to database for: "%(plug)s".',{'plug':plug})
        #correction for reading partnergroups
        if plug['plugintype'] == 'partner' and plug['isgroup']:
            plug['plugintype'] = 'partnergroep'
        #remember the plugintype
        plugintype = plug['plugintype']
            
        table = django.db.models.get_model('bots',plugintype)

        #delete fields not in model for compatibility; note that 'plugintype' is also removed.
        loopdictionary = plug.keys()
        for key in loopdictionary:
            try:
                table._meta.get_field(key)
            except django.db.models.fields.FieldDoesNotExist:
                del plug[key]

        #get key(s), put in dict 'sleutel'
        pk = table._meta.pk.name
        if pk == 'id':  #'id' is the artificial key django makes, if no key is indicated. Note the django has no 'composite keys'.
            sleutel = {}
            if table._meta.unique_together:
                for key in table._meta.unique_together[0]:
                    sleutel[key] = plug.pop(key)
        else:
            sleutel = {pk:plug.pop(pk)}

        sleutelorg = sleutel.copy()     #make a copy of the original sleutel; this is needed later
        #now we have:
        #- plugintype (is removed from plug)
        #- sleutelorg: original key fields
        #- sleutel: unique key fields. mind: translate and confirmrule have empty 'sleutel'
        #- plug: rest of database fields
        #for sleutel and plug: convert names to real database names

        #get real column names for fields in plug
        loopdictionary = plug.keys()
        for fieldname in loopdictionary:
            fieldobject = table._meta.get_field_by_name(fieldname)[0]
            try:
                if fieldobject.column != fieldname:     #if name in plug is not the real field name (in database)
                    plug[fieldobject.column] = plug[fieldname]  #add new key in plug
                    del plug[fieldname]                         #delete old key in plug
            except:
                raise botslib.PluginError(_(u'No field column for: "%(fieldname)s".'),{'fieldname':fieldname})
        #get real column names for fields in sleutel; basically the same loop but now for sleutel
        loopdictionary = sleutel.keys()
        for fieldname in loopdictionary:
            fieldobject = table._meta.get_field_by_name(fieldname)[0]
            try:
                if fieldobject.column != fieldname:
                    sleutel[fieldobject.column] = sleutel[fieldname]
                    del sleutel[fieldname]
            except:
                raise botslib.PluginError(_(u'No field column for: "%(fieldname)s".'),{'fieldname':fieldname})

        #find existing entry (if exists)
        if sleutelorg:  #note that translate and confirmrule have an empty 'sleutel'
            listexistingentries = table.objects.filter(**sleutelorg)
        elif plugintype == 'translate':
            listexistingentries = table.objects.filter(fromeditype=plug['fromeditype'],
                                                        frommessagetype=plug['frommessagetype'],
                                                        alt=plug['alt'],
                                                        frompartner=plug['frompartner_id'],
                                                        topartner=plug['topartner_id'])
        elif plugintype == 'confirmrule':
            listexistingentries = table.objects.filter(confirmtype=plug['confirmtype'],
                                                        ruletype=plug['ruletype'],
                                                        negativerule=plug['negativerule'],
                                                        idroute=plug.get('idroute'),
                                                        idchannel=plug.get('idchannel_id'),
                                                        messagetype=plug.get('messagetype'),
                                                        frompartner=plug.get('frompartner_id'),
                                                        topartner=plug.get('topartner_id'))
        if listexistingentries:
            dbobject = listexistingentries[0]  #exists, so use existing db-object
        else:
            dbobject = table(**sleutel)         #create db-object
            if plugintype == 'partner':        #for partners, first the partner needs to be saved before groups can be made
                dbobject.save()
        for key,value in plug.iteritems():      #update object with attributes from plugin
            setattr(dbobject,key,value)
        dbobject.save()                     #and save the updated object.
        botsglobal.logger.info(_(u'        Write to database is OK.'))


#*********************************************
#* plugout / make a plugin (generate)*********
#*********************************************
def make_index(cleaned_data,filename):
    ''' generate only the index file of the plugin.
        used eg for configuration change management.
    '''
    plugs = all_database2plug(cleaned_data)
    plugsasstring = make_plugs2string(plugs)
    filehandler = codecs.open(filename,'w','utf-8')
    filehandler.write(plugsasstring)
    filehandler.close()

def make_plugin(cleaned_data,filename):
    pluginzipfilehandler = zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED)

    plugs = all_database2plug(cleaned_data)
    plugsasstring = make_plugs2string(plugs)
    pluginzipfilehandler.writestr('botsindex.py',plugsasstring.encode('utf-8'))      #write index file to pluginfile
    botsglobal.logger.debug(u'    Write in index:\n %(index)s',{'index':plugsasstring})

    files4plugin = plugout_files(cleaned_data)
    for dirname, defaultdirname in files4plugin:
        pluginzipfilehandler.write(dirname,defaultdirname)
        botsglobal.logger.debug(u'    Write file "%(file)s".',{'file':defaultdirname})

    pluginzipfilehandler.close()

def all_database2plug(cleaned_data):
    ''' get all database objects, serialize these (to dict), adapt.'''
    plugs = []
    if cleaned_data['databaseconfiguration']:
        plugs += \
            database2plug(models.channel) + \
            database2plug(models.partner) + \
            database2plug(models.chanpar) + \
            database2plug(models.translate) +  \
            database2plug(models.routes) +  \
            database2plug(models.confirmrule)
    if cleaned_data['umlists']:
        plugs += \
            database2plug(models.ccodetrigger) + \
            database2plug(models.ccode)
    if cleaned_data['databasetransactions']:
        plugs += \
            database2plug(models.uniek) + \
            database2plug(models.mutex) + \
            database2plug(models.ta) + \
            database2plug(models.filereport) + \
            database2plug(models.report)
            #~ list(models.persist.objects.all()) + \       #should persist object alos be included?
    return plugs

def database2plug(db_table):
    #serialize database objects
    plugs = serializers.serialize("python", db_table.objects.all())
    if plugs:
        app,tablename = plugs[0]['model'].split('.',1)
        table = django.db.models.get_model(app,tablename)
        pk = table._meta.pk.name
        #adapt plugs
        for plug in plugs:
            plug['fields']['plugintype'] = tablename
            if pk != 'id':
                plug['fields'][pk] = plug['pk']
            #convert for correct environment: replace botssys in channels[path, mpath]
            if tablename == 'channel':
                if 'path' in plug['fields'] and plug['fields']['path'].startswith(botsglobal.ini.get('directories','botssys_org')):
                    plug['fields']['path'] = plug['fields']['path'].replace(botsglobal.ini.get('directories','botssys_org'),'botssys',1)
                if 'testpath' in plug['fields'] and plug['fields']['testpath'].startswith(botsglobal.ini.get('directories','botssys_org')):
                    plug['fields']['testpath'] = plug['fields']['testpath'].replace(botsglobal.ini.get('directories','botssys_org'),'botssys',1)
    return plugs

def make_plugs2string(plugs):
    ''' return plugs (serialized objects) as unicode strings.
    '''
    lijst = [u'# -*- coding: utf-8 -*-',u'import datetime',"version = '%s'" % (botsglobal.version),'plugins = [']
    lijst.extend([plug2string(plug['fields']) for plug in plugs])
    lijst.append(u']\n')
    return '\n'.join(lijst)

def plug2string(plugdict):
    ''' like repr() for a dict, but:
        - starts with 'plugintype'
        - other entries are sorted; this because of predictability
        - produce unicode by using str().decode(unicode_escape): bytes->unicode; converts escaped unicode-chrs to correct unicode. repr produces these.
        str().decode(): bytes->unicode
        str().encode(): unicode->bytes
    '''
    terug = u"{" + repr('plugintype') + u": " + repr(plugdict.pop('plugintype'))
    for key in sorted(plugdict):
        terug += u", " + repr(key) + u": " + repr(plugdict[key])
    terug += u'},'
    return terug

def plugout_files(cleaned_data):
    ''' gather list of files for the plugin that is generated.
    '''
    files2return = []
    usersys = unicode(botsglobal.ini.get('directories','usersysabs'))
    botssys = unicode(botsglobal.ini.get('directories','botssys'))
    if cleaned_data['fileconfiguration']:       #gather from usersys
        files2return.extend(plugout_files_bydir(usersys,u'usersys'))
        if not cleaned_data['charset']:     #if edifact charsets are not needed: remove them (are included in default bots installation).
            charsetdirs = plugout_files_bydir(os.path.join(usersys,u'charsets'),u'usersys/charsets')
            for charset in charsetdirs:
                try:
                    index = files2return.index(charset)
                    files2return.pop(index)
                except ValueError:
                    pass
    else:
        if cleaned_data['charset']:     #if edifact charsets are not needed: remove them (are included in default bots installation).
            files2return.extend(plugout_files_bydir(os.path.join(usersys,u'charsets'),u'usersys/charsets'))
    if cleaned_data['config']:
        config = botsglobal.ini.get('directories','config')
        files2return.extend(plugout_files_bydir(config,u'config'))
    if cleaned_data['data']:
        data = botsglobal.ini.get('directories','data')
        files2return.extend(plugout_files_bydir(data,u'botssys/data'))
    if cleaned_data['database']:
        files2return.extend(plugout_files_bydir(os.path.join(botssys,u'sqlitedb'),u'botssys/sqlitedb.copy'))  #yeah...readign a plugin with a new database will cause a crash...do this manually...
    if cleaned_data['infiles']:
        files2return.extend(plugout_files_bydir(os.path.join(botssys,u'infile'),u'botssys/infile'))
    if cleaned_data['logfiles']:
        log_file = botsglobal.ini.get('directories','logging')
        files2return.extend(plugout_files_bydir(log_file,u'botssys/logging'))
    return files2return

def plugout_files_bydir(dirname,defaultdirname):
    ''' gather all files from directory dirname'''
    files2return = []
    for root, dirs, files in os.walk(dirname):
        head, tail = os.path.split(root)
        #convert for correct environment: replace dirname with the default directory name
        rootinplugin = root.replace(dirname,defaultdirname,1)
        for bestand in files:
            ext = os.path.splitext(bestand)[1]
            if ext and (ext in ['.pyc','.pyo'] or bestand in ['__init__.py']):
                continue
            files2return.append([os.path.join(root,bestand),os.path.join(rootinplugin,bestand)])
    return files2return




''' Declare database tabels. 
    Django is not always perfect in generating db - but improving ;-)). 
    The generated database can be manipulated SQL. see bots/sql/*.
'''
import os
import urllib
import re
from django.db import models
from django.utils.translation import ugettext_lazy as _     #djnago 1.7: have to use ugettext_lazy here
#~ from django.core.validators import validate_email
from django.core.validators import validate_integer
from django.core.exceptions import ValidationError
import botsglobal
import validate_email
#***Declare constants, mostly codelists.**********************************************
DEFAULT_ENTRY = ('',"---------")
STATUST = [
    (0, _(u'Open')),
    (1, _(u'Error')),
    (2, _(u'Stuck')),
    (3, _(u'Done')),
    (4, _(u'Resend')),
    (5, _(u'No retry')),
    ]
STATUS = [
    (1,_(u'Process')),
    (3,_(u'Discarded')),
    (200,_(u'Received')),
    (220,_(u'Infile')),
    (310,_(u'Parsed')),
    (320,_(u'Document-in')),
    (330,_(u'Document-out')),
    (400,_(u'Merged')),
    (500,_(u'Outfile')),
    (520,_(u'Send')),
    ]
EDITYPES = [
    #~ DEFAULT_ENTRY,
    ('csv', _(u'csv')),
    ('database', _(u'database (old)')),
    ('db', _(u'db')),
    ('edifact', _(u'edifact')),
    ('email-confirmation',_(u'email-confirmation')),
    ('excel',_(u'excel (only incoming)')),
    ('fixed', _(u'fixed')),
    ('idoc', _(u'idoc')),
    ('json', _(u'json')),
    ('jsonnocheck', _(u'jsonnocheck')),
    ('mailbag', _(u'mailbag')),
    ('raw', _(u'raw')),
    ('templatehtml', _(u'template-html')),
    ('tradacoms', _(u'tradacoms')),
    ('xml', _(u'xml')),
    ('xmlnocheck', _(u'xmlnocheck')),
    ('x12', _(u'x12')),
    ]
INOROUT = (
    ('in', _(u'in')),
    ('out', _(u'out')),
    )
CHANNELTYPE = (     #Note: in communication.py these channeltypes are converted to channeltype to use in acceptance tests.
    ('file', _(u'file')),
    ('smtp', _(u'smtp')),
    ('smtps', _(u'smtps')),
    ('smtpstarttls', _(u'smtpstarttls')),
    ('pop3', _(u'pop3')),
    ('pop3s', _(u'pop3s')),
    ('pop3apop', _(u'pop3apop')),
    ('http', _(u'http')),
    ('https', _(u'https')),
    ('imap4', _(u'imap4')),
    ('imap4s', _(u'imap4s')),
    ('ftp', _(u'ftp')),
    ('ftps', _(u'ftps (explicit)')),
    ('ftpis', _(u'ftps (implicit)')),
    ('sftp', _(u'sftp (ssh)')),
    ('xmlrpc', _(u'xmlrpc')),
    ('mimefile', _(u'mimefile')),
    ('trash', _(u'trash/discard')),
    ('communicationscript', _(u'communicationscript')),
    ('db', _(u'db')),
    ('database', _(u'database (old)')),
    )
CONFIRMTYPE = [
    ('ask-email-MDN',_(u'ask an email confirmation (MDN) when sending')),
    ('send-email-MDN',_(u'send an email confirmation (MDN) when receiving')),
    ('ask-x12-997',_(u'ask a x12 confirmation (997) when sending')),
    ('send-x12-997',_(u'send a x12 confirmation (997) when receiving')),
    ('ask-edifact-CONTRL',_(u'ask an edifact confirmation (CONTRL) when sending')),
    ('send-edifact-CONTRL',_(u'send an edifact confirmation (CONTRL) when receiving')),
    ]
RULETYPE = (
    ('all',_(u'Confirm all')),
    ('route',_(u'Route')),
    ('channel',_(u'Channel')),
    ('frompartner',_(u'Frompartner')),
    ('topartner',_(u'Topartner')),
    ('messagetype',_(u'Messagetype')),
    )
ENCODE_MIME = (
    ('always',_(u'Base64')),
    ('never',_(u'Never')),
    ('ascii',_(u'Base64 if not ascii')),
    )
EDI_AS_ATTACHMENT = (
    ('attachment',_(u'As attachment')),
    ('body',_(u'In body of email')),
    )
ENCODE_ZIP_IN = (
    (1,_(u'Always unzip file')),
    (2,_(u'If zip-file: unzip')),
    )
ENCODE_ZIP_OUT = (
    (1,_(u'Always zip')),
    )
TRANSLATETYPES = (
    (0,_(u'Nothing')),
    (1,_(u'Translate')),
    (2,_(u'Pass-through')),
    (3,_(u'Parse & Pass-through')),
    )
CONFIRMTYPELIST = [DEFAULT_ENTRY] + CONFIRMTYPE
EDITYPESLIST = [DEFAULT_ENTRY] + EDITYPES

#***Functions that produced codelists.**********************************************
def getroutelist():     #needed because the routeid is needed (and this is not theprimary key
    return [DEFAULT_ENTRY]+[(l,l) for l in routes.objects.values_list('idroute', flat=True).order_by('idroute').distinct() ]

def getinmessagetypes():
    return [DEFAULT_ENTRY]+[(l,l) for l in translate.objects.values_list('frommessagetype', flat=True).order_by('frommessagetype').distinct() ]
def getoutmessagetypes():
    return [DEFAULT_ENTRY]+[(l,l) for l in translate.objects.values_list('tomessagetype', flat=True).order_by('tomessagetype').distinct() ]
def getallmessagetypes():
    return [DEFAULT_ENTRY]+[(l,l) for l in sorted(set(list(translate.objects.values_list('tomessagetype', flat=True).all()) + list(translate.objects.values_list('frommessagetype', flat=True).all()) )) ]
def getpartners():
    return [DEFAULT_ENTRY]+[(l,'%s (%s)'%(l,n)) for (l,n) in partner.objects.values_list('idpartner','name').filter(active=True)]
def getfromchannels():
    return [DEFAULT_ENTRY]+[(l,'%s (%s)'%(l,t)) for (l,t) in channel.objects.values_list('idchannel','type').filter(inorout='in')]
def gettochannels():
    return [DEFAULT_ENTRY]+[(l,'%s (%s)'%(l,t)) for (l,t) in channel.objects.values_list('idchannel','type').filter(inorout='out')]

#***Database tables that produced codelists.**********************************************
class StripCharField(models.CharField):
    ''' strip values before saving to database. this is not default in django #%^&*'''
    def get_prep_value(self, value,*args,**kwargs):
        ''' Convert Python objects (value) to query values (returned)
        ''' 
        if isinstance(value, basestring):
            return value.strip()
        else:
            return value

def multiple_email_validator(value):
    ''' Problems with validating email adresses:
        django's email validating is to strict. (eg if quoted user part, space is not allowed).
        use case: x400 via IPmail (x400 addresses are used in email-addresses).
        Use re-expressions to get this better/conform email standards.
    '''
    if botsglobal.ini.getboolean('webserver','use_email_address_validation',True):      #tric to disable email validation via bots.ini
        emails = re.split(',(?=(?:[^\"]*\"[^\"]*\")*[^\"]*$)',value)    #split emails
        for email in emails:
            if not validate_email.validate_email_address(email):
                raise ValidationError(_(u'Enter valid e-mail address(es) separated by commas.'), code='invalid')

def script_link1(script,linktext):
    ''' if script exists return a plain text name as link; else return "no" icon, plain text name
        used in translate (all scripts should exist, missing script is an error).
    '''
    if os.path.exists(script):
        return '<a href="/srcfiler/?src=%s" target="_blank">%s</a>'%(urllib.quote(script.encode("utf-8")),linktext)
    else:
        return '<img src="/media/admin/img/icon-no.gif"></img> %s'%linktext

def script_link2(script):
    ''' if script exists return "yes" icon + view link; else return "no" icon
        used in routes, channels (scripts are optional)
    '''
    if os.path.exists(script):
        return '<a class="nowrap" href="/srcfiler/?src=%s" target="_blank"><img src="/media/admin/img/icon-yes.gif"></img> view</a>'%urllib.quote(script.encode("utf-8"))
    else:
        return '<img src="/media/admin/img/icon-no.gif"></img>'


class MultipleEmailField(models.CharField):
    default_validators = [multiple_email_validator]
    description = _('One or more e-mail address(es),separated by ",".')
class TextAsInteger(models.CharField):
    default_validators = [validate_integer]

#***********************************************************************************
#******** written by webserver ********************************************************
#***********************************************************************************
class confirmrule(models.Model):
    #~ id = models.IntegerField(primary_key=True)
    active = models.BooleanField(default=False)
    confirmtype = StripCharField(max_length=35,choices=CONFIRMTYPE)
    ruletype = StripCharField(max_length=35,choices=RULETYPE)
    negativerule = models.BooleanField(default=False,help_text=_(u'Use to exclude. Bots first checks positive rules, than negative rules. Eg include certain channel, exclude partner XXX.'))
    frompartner = models.ForeignKey('partner',related_name='cfrompartner',null=True,on_delete=models.CASCADE,blank=True,limit_choices_to = {'isgroup': False})
    topartner = models.ForeignKey('partner',related_name='ctopartner',null=True,on_delete=models.CASCADE,blank=True,limit_choices_to = {'isgroup': False})
    idroute = StripCharField(max_length=35,null=True,blank=True,verbose_name=_(u'route'))
    idchannel = models.ForeignKey('channel',null=True,on_delete=models.CASCADE,blank=True,verbose_name=_(u'channel'))
    editype = StripCharField(max_length=35,choices=EDITYPES,blank=True)         #20121229"is not used anymore.....editype is always clear from context.
    messagetype = StripCharField(max_length=35,blank=True,help_text=_(u'Eg "850004010" (x12) or "ORDERSD96AUN" (edifact).'))
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501
    rsrv2 = models.IntegerField(null=True)                        #added 20100501
    def __unicode__(self):
        return unicode(self.confirmtype) + u' ' + unicode(self.ruletype)
    class Meta:
        db_table = 'confirmrule'
        verbose_name = _(u'confirm rule')
        ordering = ['confirmtype','ruletype','negativerule','frompartner','topartner','idroute','idchannel','messagetype']
class ccodetrigger(models.Model):
    ccodeid = StripCharField(primary_key=True,max_length=35,verbose_name=_(u'Type of user code'))
    ccodeid_desc = models.TextField(blank=True,null=True,verbose_name=_(u'Description'))
    def __unicode__(self):
        return unicode(self.ccodeid)
    class Meta:
        db_table = 'ccodetrigger'
        verbose_name = _(u'user code type')
        ordering = ['ccodeid']
class ccode(models.Model):
    #~ id = models.IntegerField(primary_key=True)     #added 20091221
    ccodeid = models.ForeignKey(ccodetrigger,on_delete=models.CASCADE,verbose_name=_(u'Type of user code'))
    leftcode = StripCharField(max_length=35,db_index=True)
    rightcode = StripCharField(max_length=70,db_index=True)
    attr1 = StripCharField(max_length=70,blank=True)
    attr2 = StripCharField(max_length=35,blank=True)
    attr3 = StripCharField(max_length=35,blank=True)
    attr4 = StripCharField(max_length=35,blank=True)
    attr5 = StripCharField(max_length=35,blank=True)
    attr6 = StripCharField(max_length=35,blank=True)
    attr7 = StripCharField(max_length=35,blank=True)
    attr8 = StripCharField(max_length=35,blank=True)
    def __unicode__(self):
        return unicode(self.ccodeid) + u' ' + unicode(self.leftcode) + u' ' + unicode(self.rightcode)
    class Meta:
        db_table = 'ccode'
        verbose_name = _(u'user code')
        unique_together = (('ccodeid','leftcode','rightcode'),)
        ordering = ['ccodeid','leftcode']
class channel(models.Model):
    idchannel = StripCharField(max_length=35,primary_key=True)
    inorout = StripCharField(max_length=35,choices=INOROUT,verbose_name=_(u'in/out'))
    type = StripCharField(max_length=35,choices=CHANNELTYPE)        #protocol type: ftp, smtp, file, etc
    charset = StripCharField(max_length=35,default=u'us-ascii')     #20120828: not used anymore; in database is NOT NULL
    host = StripCharField(max_length=256,blank=True)
    port = models.PositiveIntegerField(default=0,blank=True,null=True)
    username = StripCharField(max_length=35,blank=True)
    secret = StripCharField(max_length=35,blank=True,verbose_name=_(u'password'))
    starttls = models.BooleanField(default=False,verbose_name='No check from-address',help_text=_(u"Do not check if incoming 'from' email addresses is known."))       #20091027: used as 'no check on "from:" email address'
    apop = models.BooleanField(default=False,verbose_name='No check to-address',help_text=_(u"Do not check if incoming 'to' email addresses is known."))       #20110104: used as 'no check on "to:" email address'
    remove = models.BooleanField(default=False,help_text=_(u"Delete incoming edi files after reading.<br>Use in production else files are read again and again."))
    path = StripCharField(max_length=256,blank=True)  #different from host - in ftp both host and path are used
    filename = StripCharField(max_length=256,blank=True,help_text=_(u'Incoming: use wild-cards eg: "*.edi".<br>Outgoing: many options, see <a target="_blank" href="http://code.google.com/p/bots/wiki/Filenames">wiki</a>.<br>Advised: use "*" in filename (is replaced by unique counter per channel).<br>eg "D_*.edi" gives D_1.edi, D_2.edi, etc.'))
    lockname = StripCharField(max_length=35,blank=True,verbose_name=_(u'Lock-file'),help_text=_(u'Directory locking: if lock-file exists in directory, directory is locked for reading/writing.'))
    syslock = models.BooleanField(default=False,verbose_name=_(u'System locks'),help_text=_(u'Use system file locks for reading or writing edi files (windows, *nix).'))
    parameters = StripCharField(max_length=70,blank=True,help_text=_(u'For use in user communication scripting.'))
    ftpaccount = StripCharField(max_length=35,blank=True,verbose_name=_(u'ftp account'),help_text=_(u'FTP account information; note that few systems implement this.'))
    ftpactive = models.BooleanField(default=False,verbose_name=_(u'ftp active mode'),help_text=_(u'Passive mode is used unless this is ticked.'))
    ftpbinary = models.BooleanField(default=False,verbose_name=_(u'ftp binary transfer mode'),help_text=_(u'File transfers are ASCII unless this is ticked.'))
    askmdn = StripCharField(max_length=17,blank=True,choices=ENCODE_MIME,verbose_name=_(u'mime encoding'))     #20100703: used to indicate mime-encoding
    sendmdn = StripCharField(max_length=17,blank=True,choices=EDI_AS_ATTACHMENT,verbose_name=_(u'as body or attachment'))      #20120922: for email/mime: edi file as attachment or in body 
    mdnchannel = StripCharField(max_length=35,blank=True,verbose_name=_(u'Tmp-part file name'),help_text=_(u'Write file than rename. Bots renames to filename without this tmp-part.<br>Eg first write "myfile.edi.tmp", tmp-part is ".tmp", rename to "myfile.edi".'))      #20140113:use as tmp part of file name
    archivepath = StripCharField(max_length=256,blank=True,verbose_name=_(u'Archive path'),help_text=_(u'Write edi files to an archive.<br>See <a target="_blank" href="http://code.google.com/p/bots/wiki/Archiving">wiki</a>. Eg: "C:/edi/archive/mychannel".'))           #added 20091028
    desc = models.TextField(max_length=256,null=True,blank=True,verbose_name=_(u'Description'))
    rsrv1 = TextAsInteger(max_length=35,blank=True,null=True,verbose_name=_(u'Max failures'),help_text=_(u'Max number of connection failures of incommunication before this is reported as a processerror (default: direct report).'))      #added 20100501 #20140315: used as max_com
    rsrv2 = models.IntegerField(null=True,blank=True,verbose_name=_(u'Max seconds'),help_text=_(u'Max seconds for in-communication channel to run. Purpose: limit incoming edi files; for large volumes it is better read more often than all files in one time.'))   #added 20100501. 20110906: max communication time.
    rsrv3 = models.IntegerField(null=True,blank=True,verbose_name=_(u'Max days archive'),help_text=_(u'Max number of days files are kept in archive.<br>Overrules global setting in bots.ini.'))   #added 20121030. #20131231: use as maxdaysarchive
    keyfile = StripCharField(max_length=256,blank=True,null=True,verbose_name=_(u'Private key file'),help_text=_(u'Path to file that contains PEM formatted private key.'))          #added 20121201
    certfile = StripCharField(max_length=256,blank=True,null=True,verbose_name=_(u'Certificate chain file'),help_text=_(u'Path to file that contains PEM formatted certificate chain.'))          #added 20121201
    testpath = StripCharField(max_length=256,blank=True,verbose_name=_(u'Acceptance test path'),help_text=_(u'Path used during acceptance tests, see <a target="_blank" href="http://code.google.com/p/bots/wiki/DeploymentAcceptance">wiki</a>.'))           #added 20120111

    def communicationscript(self):
        return script_link2(os.path.join(botsglobal.ini.get('directories','usersysabs'),'communicationscripts', self.idchannel + '.py'))
    communicationscript.allow_tags = True
    communicationscript.short_description = 'User script'

    class Meta:
        ordering = ['idchannel']
        db_table = 'channel'
    def __unicode__(self):
        return self.idchannel + u' (' + self.type + u')'
class partner(models.Model):
    idpartner = StripCharField(max_length=35,primary_key=True,verbose_name=_(u'partner identification'))
    active = models.BooleanField(default=False)
    isgroup = models.BooleanField(default=False,help_text=_(u'Indicate if normal partner or a partner group. Partners can be assigned to partner groups.'))
    name = StripCharField(max_length=256) #only used for user information
    mail = MultipleEmailField(max_length=256,blank=True)
    cc = MultipleEmailField(max_length=256,blank=True,help_text=_(u'Multiple CC-addresses supported (comma-separated).'))
    mail2 = models.ManyToManyField(channel, through='chanpar',blank=True)
    group = models.ManyToManyField("self",db_table='partnergroup',blank=True,symmetrical=False,limit_choices_to = {'isgroup': True})
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501
    rsrv2 = models.IntegerField(null=True)                        #added 20100501
    name1 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    name2 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    name3 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    address1 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    address2 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    address3 = StripCharField(max_length=70,blank=True,null=True)          #added 20121201
    city = StripCharField(max_length=35,blank=True,null=True)          #added 20121201
    postalcode = StripCharField(max_length=17,blank=True,null=True)          #added 20121201
    countrysubdivision = StripCharField(max_length=9,blank=True,null=True)          #added 20121201
    countrycode = StripCharField(max_length=3,blank=True,null=True)          #added 20121201
    phone1 = StripCharField(max_length=17,blank=True,null=True)          #added 20121201
    phone2 = StripCharField(max_length=17,blank=True,null=True)          #added 20121201
    startdate = models.DateField(blank=True,null=True)          #added 20121201
    enddate = models.DateField(blank=True,null=True)          #added 20121201
    desc = models.TextField(blank=True,null=True,verbose_name=_(u'Description'))    #added 20121201
    attr1 = StripCharField(max_length=35,blank=True,null=True,verbose_name=_(u'attr1')) # user can customise verbose name
    attr2 = StripCharField(max_length=35,blank=True,null=True,verbose_name=_(u'attr2'))
    attr3 = StripCharField(max_length=35,blank=True,null=True,verbose_name=_(u'attr3'))
    attr4 = StripCharField(max_length=35,blank=True,null=True,verbose_name=_(u'attr4'))
    attr5 = StripCharField(max_length=35,blank=True,null=True,verbose_name=_(u'attr5'))
    class Meta:
        ordering = ['idpartner']
        db_table = 'partner'
    def __unicode__(self):
        return unicode(self.idpartner) + u' (' + unicode(self.name) + u')'
    def save(self, *args, **kwargs):
        if isinstance(self,partnergroep):
            self.isgroup = True
        else:
            self.isgroup = False
        super(partner,self).save(*args,**kwargs)
        
class partnergroep(partner):
    class Meta:
        proxy = True
        ordering = ['idpartner']
        db_table = 'partner'

class chanpar(models.Model):
    #~ id = models.IntegerField(primary_key=True)     #added 20091221
    idpartner = models.ForeignKey(partner,on_delete=models.CASCADE,verbose_name=_(u'partner'))
    idchannel = models.ForeignKey(channel,on_delete=models.CASCADE,verbose_name=_(u'channel'))
    mail = MultipleEmailField(max_length=256)
    cc = MultipleEmailField(max_length=256,blank=True)           #added 20091111
    askmdn = models.BooleanField(default=False)     #not used anymore 20091019
    sendmdn = models.BooleanField(default=False)    #not used anymore 20091019
    class Meta:
        unique_together = (('idpartner','idchannel'),)
        ordering = ['idpartner','idchannel']
        db_table = 'chanpar'
        verbose_name = _(u'email address per channel')
        verbose_name_plural = _(u'email address per channel')
    def __unicode__(self):
        return unicode(self.idpartner) + u' ' + unicode(self.idchannel) + u' ' + unicode(self.mail)
class translate(models.Model):
    #~ id = models.IntegerField(primary_key=True)
    active = models.BooleanField(default=False)
    fromeditype = StripCharField(max_length=35,choices=EDITYPES,help_text=_(u'Editype to translate from.'))
    frommessagetype = StripCharField(max_length=35,help_text=_(u'Messagetype to translate from.'))
    alt = StripCharField(max_length=35,null=False,blank=True,verbose_name=_(u'Alternative translation'),help_text=_(u'Do translation only for this alternative translation.'))
    frompartner = models.ForeignKey(partner,related_name='tfrompartner',null=True,blank=True,on_delete=models.PROTECT,help_text=_(u'Do translation only for this frompartner.'))
    topartner = models.ForeignKey(partner,related_name='ttopartner',null=True,blank=True,on_delete=models.PROTECT,help_text=_(u'Do translation only for this topartner.'))
    tscript = StripCharField(max_length=35,verbose_name=_(u'Mapping Script'),help_text=_(u'Mappingscript to use in translation.'))
    toeditype = StripCharField(max_length=35,choices=EDITYPES,help_text=_(u'Editype to translate to.'))
    tomessagetype = StripCharField(max_length=35,help_text=_(u'Messagetype to translate to.'))
    desc = models.TextField(max_length=256,null=True,blank=True,verbose_name=_(u'Description'))
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501
    rsrv2 = models.IntegerField(null=True)                        #added 20100501

    def tscript_link(self):
        return script_link1(os.path.join(botsglobal.ini.get('directories','usersysabs'),'mappings', self.fromeditype, self.tscript + '.py'),self.tscript)
    tscript_link.allow_tags = True
    tscript_link.short_description = 'Mapping Script'

    def frommessagetype_link(self):
        return script_link1(os.path.join(botsglobal.ini.get('directories','usersysabs'),'grammars', self.fromeditype, self.frommessagetype + '.py'),self.frommessagetype)
    frommessagetype_link.allow_tags = True
    frommessagetype_link.short_description = 'Frommessagetype'

    def tomessagetype_link(self):
        return script_link1(os.path.join(botsglobal.ini.get('directories','usersysabs'),'grammars', self.toeditype, self.tomessagetype + '.py'),self.tomessagetype)
    tomessagetype_link.allow_tags = True
    tomessagetype_link.short_description = 'Tomessagetype'

    class Meta:
        db_table = 'translate'
        verbose_name = _(u'translation rule')
        ordering = ['fromeditype','frommessagetype','frompartner','topartner','alt']
    def __unicode__(self):
        return unicode(self.fromeditype) + u' ' + unicode(self.frommessagetype) + u' ' + unicode(self.alt) + u' ' + unicode(self.frompartner) + u' ' + unicode(self.topartner)


class routes(models.Model):
    #~ id = models.IntegerField(primary_key=True)
    idroute = StripCharField(max_length=35,db_index=True,help_text=_(u'Identification of route; a composite route consists of multiple parts having the same "idroute".'))
    seq = models.PositiveIntegerField(default=1,verbose_name=_(u'Sequence'),help_text=_(u'For routes consisting of multiple parts, this indicates the order these parts are run.'))
    active = models.BooleanField(default=False,help_text=_(u'Bots-engine only uses active routes.'))
    fromchannel = models.ForeignKey(channel,related_name='rfromchannel',null=True,on_delete=models.SET_NULL,blank=True,verbose_name=_(u'incoming channel'),limit_choices_to = {'inorout': 'in'},help_text=_(u'Receive edi files via this communication channel.'))
    fromeditype = StripCharField(max_length=35,choices=EDITYPES,blank=True,help_text=_(u'Editype of the incoming edi files.'))
    frommessagetype = StripCharField(max_length=35,blank=True,help_text=_(u'Messagetype of incoming edi files. For edifact: messagetype=edifact; for x12: messagetype=x12.'))
    tochannel = models.ForeignKey(channel,related_name='rtochannel',null=True,on_delete=models.SET_NULL,blank=True,verbose_name=_(u'outgoing channel'),limit_choices_to = {'inorout': 'out'},help_text=_(u'Send edi files via this communication channel.'))
    toeditype = StripCharField(max_length=35,choices=EDITYPES,blank=True,help_text=_(u'Filter edi files of this editype for outgoing channel.'))
    tomessagetype = StripCharField(max_length=35,blank=True,help_text=_(u'Filter edi files of this messagetype for outgoing channel.'))
    alt = StripCharField(max_length=35,default=u'',blank=True,verbose_name='Alternative translation',help_text=_(u'Only use if there is more than one "translation" for the same editype and messagetype.'))
    frompartner = models.ForeignKey(partner,related_name='rfrompartner',null=True,on_delete=models.SET_NULL,blank=True,limit_choices_to = {'isgroup': False},help_text=_(u'The frompartner of the incoming edi files.'))
    topartner = models.ForeignKey(partner,related_name='rtopartner',null=True,on_delete=models.SET_NULL,blank=True,limit_choices_to = {'isgroup': False},help_text=_(u'The topartner of the incoming edi files.'))
    frompartner_tochannel = models.ForeignKey(partner,related_name='rfrompartner_tochannel',null=True,on_delete=models.PROTECT,blank=True,help_text=_(u'Filter edi files of this partner/partnergroup for outgoing channel'))
    topartner_tochannel = models.ForeignKey(partner,related_name='rtopartner_tochannel',null=True,on_delete=models.PROTECT,blank=True,help_text=_(u'Filter edi files of this partner/partnergroup for outgoing channel'))
    testindicator = StripCharField(max_length=1,blank=True,help_text=_(u'Filter edi files with this test-indicator for outgoing channel.'))
    translateind = models.IntegerField(default=1,choices=TRANSLATETYPES,verbose_name='translate',help_text=_(u'Indicates what to do with incoming files for this route(part).'))
    notindefaultrun = models.BooleanField(default=False,blank=True,verbose_name=_(u'Not in default run'),help_text=_(u'Do not use this route in a normal run. Advanced, related to scheduling specific routes or not.'))
    desc = models.TextField(max_length=256,null=True,blank=True,verbose_name=_(u'Description'))
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501 
    rsrv2 = models.IntegerField(null=True,blank=True)           #added 20100501
    defer = models.BooleanField(default=False,blank=True,help_text=_(u'Set ready for communication, defer actual communication. Communication is done later in another route(-part).'))                        #added 20100601
    zip_incoming = models.IntegerField(null=True,blank=True,choices=ENCODE_ZIP_IN,verbose_name=_(u'Incoming zip-file handling'),help_text=_(u'Unzip received files.'))  #added 20100501 #20120828: use for zip-options
    zip_outgoing = models.IntegerField(null=True,blank=True,choices=ENCODE_ZIP_OUT,verbose_name=_(u'Outgoing zip-file handling'),help_text=_(u'Send files as zip-files.'))                        #added 20100501
    
    def routescript(self):
        return script_link2(os.path.join(botsglobal.ini.get('directories','usersysabs'),'routescripts', self.idroute + '.py'))
    routescript.allow_tags = True
    routescript.short_description = 'Script'
    
    def indefaultrun(obj):
        return not obj.notindefaultrun
    indefaultrun.boolean = True
    indefaultrun.short_description = 'Default run'
    
    class Meta:
        db_table = 'routes'
        verbose_name = _(u'route')
        unique_together = (('idroute','seq'),)
        ordering = ['idroute','seq']
    def __unicode__(self):
        return unicode(self.idroute) + u' ' + unicode(self.seq)
    def translt(self):
        if self.translateind == 0:
            return '<img alt="%s" src="/media/admin/img/icon-no.gif"></img>'%(self.get_translateind_display())
        elif self.translateind == 1:
            return '<img alt="%s" src="/media/admin/img/icon-yes.gif"></img>'%(self.get_translateind_display())
        elif self.translateind == 2:
            return '<img alt="%s" src="/media/images/icon-pass.gif"></img>'%(self.get_translateind_display())
        elif self.translateind == 3:
            return '<img alt="%s" src="/media/images/icon-pass_parse.gif"></img>'%(self.get_translateind_display())
    translt.allow_tags = True
    translt.admin_order_field = 'translateind'

#***********************************************************************************
#******** written by engine ********************************************************
#***********************************************************************************
class filereport(models.Model):
    #~ id = models.IntegerField(primary_key=True)
    idta = models.IntegerField(primary_key=True)
    reportidta = models.IntegerField()
    statust = models.IntegerField(choices=STATUST)
    retransmit = models.IntegerField()
    idroute = StripCharField(max_length=35)
    fromchannel = StripCharField(max_length=35)
    tochannel = StripCharField(max_length=35)
    frompartner = StripCharField(max_length=35)
    topartner = StripCharField(max_length=35)
    frommail = StripCharField(max_length=256)
    tomail = StripCharField(max_length=256)
    ineditype = StripCharField(max_length=35,choices=EDITYPES)
    inmessagetype = StripCharField(max_length=35)
    outeditype = StripCharField(max_length=35,choices=EDITYPES)
    outmessagetype = StripCharField(max_length=35)
    incontenttype = StripCharField(max_length=35)
    outcontenttype = StripCharField(max_length=35)
    nrmessages = models.IntegerField()
    ts = models.DateTimeField(db_index=True) #copied from ta
    infilename = StripCharField(max_length=256)
    inidta = models.IntegerField(null=True)   #not used anymore
    outfilename = StripCharField(max_length=256)
    outidta = models.IntegerField()
    divtext = StripCharField(max_length=35)
    errortext = models.TextField()
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501; 20120618: email subject
    rsrv2 = models.IntegerField(null=True)                        #added 20100501
    filesize = models.IntegerField(null=True)                    #added 20121030
    class Meta:
        db_table = 'filereport'
class mutex(models.Model):
    #specific SQL is used (database defaults are used)
    mutexk = models.IntegerField(primary_key=True)  #is always value '1'
    mutexer = models.IntegerField() 
    ts = models.DateTimeField()         #timestamp of mutex
    class Meta:
        db_table = 'mutex'
class persist(models.Model):
    #OK, this has gone wrong. There is no primary key here, so django generates this. But there is no ID in the custom sql.
    #Django still uses the ID in sql manager. This leads to an error in snapshot plugin. Disabled this in snapshot function; to fix this really database has to be changed.
    #specific SQL is used (database defaults are used)
    domein = StripCharField(max_length=35)
    botskey = StripCharField(max_length=35)
    content = models.TextField()
    ts = models.DateTimeField()
    class Meta:
        db_table = 'persist'
        unique_together = (("domein","botskey"),)
class report(models.Model):
    idta = models.IntegerField(primary_key=True)    #rename to reportidta
    lastreceived = models.IntegerField()
    lastdone = models.IntegerField()
    lastopen = models.IntegerField()
    lastok = models.IntegerField()
    lasterror = models.IntegerField()
    send = models.IntegerField()
    processerrors = models.IntegerField()
    ts = models.DateTimeField(db_index=True)                     #copied from (runroot)ta
    type = StripCharField(max_length=35)
    status = models.BooleanField()
    rsrv1 = StripCharField(max_length=35,blank=True,null=True)  #added 20100501. 20131230: used to store the commandline for the run.
    rsrv2 = models.IntegerField(null=True)                       #added 20100501.
    filesize = models.IntegerField(null=True)                    #added 20121030: total size of messages that have been translated.
    acceptance = models.IntegerField(null=True)                            #added 20130114: 
    class Meta:
        db_table = 'report'
#~ #trigger for sqlite to use local time (instead of utc). I can not add this to sqlite specific sql code, as django does not allow complex (begin ... end) sql here.
#~ CREATE TRIGGER uselocaltime  AFTER INSERT ON ta
#~ BEGIN
#~ UPDATE ta
#~ SET ts = datetime('now','localtime')
#~ WHERE idta = new.idta ;
#~ END;
class ta(models.Model):
    #specific SQL is used (database defaults are used)
    idta = models.AutoField(primary_key=True)
    statust = models.IntegerField(choices=STATUST)
    status = models.IntegerField(choices=STATUS)
    parent = models.IntegerField(db_index=True)
    child = models.IntegerField()
    script = models.IntegerField()
    idroute = StripCharField(max_length=35)
    filename = StripCharField(max_length=256)
    frompartner = StripCharField(max_length=35)
    topartner = StripCharField(max_length=35)
    fromchannel = StripCharField(max_length=35)
    tochannel = StripCharField(max_length=35)
    editype = StripCharField(max_length=35)
    messagetype = StripCharField(max_length=35)
    alt = StripCharField(max_length=35)
    divtext = StripCharField(max_length=35)             #name of translation script. 
    merge = models.BooleanField()
    nrmessages = models.IntegerField()
    testindicator = StripCharField(max_length=10)     #0:production; 1:test. Length to 1?
    reference = StripCharField(max_length=70,db_index=True)
    frommail = StripCharField(max_length=256)
    tomail = StripCharField(max_length=256)
    charset = StripCharField(max_length=35)
    statuse = models.IntegerField()                     #obsolete 20091019
    retransmit = models.BooleanField()                  #20070831: only retransmit, not rereceive
    contenttype = StripCharField(max_length=35)
    errortext = models.TextField()                    #20120921: unlimited length
    ts = models.DateTimeField()
    confirmasked = models.BooleanField()              #added 20091019; confirmation asked or send
    confirmed = models.BooleanField()                 #added 20091019; is confirmation received (when asked)
    confirmtype = StripCharField(max_length=35)       #added 20091019
    confirmidta = models.IntegerField()               #added 20091019
    envelope = StripCharField(max_length=35)          #added 20091024
    botskey = StripCharField(max_length=35)           #added 20091024
    cc = StripCharField(max_length=512)               #added 20091111
    rsrv1 = StripCharField(max_length=35)             #added 20100501; 20120618: email subject
    rsrv2 = models.IntegerField(null=True)            #added 20100501;
    rsrv3 = StripCharField(max_length=35)             #added 20100501; 20131231: envelopeID to explicitly control enveloping (enveloping criterium)
    rsrv4 = models.IntegerField(null=True)            #added 20100501; 
    rsrv5 = StripCharField(max_length=35)             #added 20121030
    filesize = models.IntegerField(null=True)         #added 20121030; 
    numberofresends = models.IntegerField(null=True)  #added 20121030; if all OK (no resend) this is 0
    class Meta:
        db_table = 'ta'
class uniek(models.Model):
    #specific SQL is used (database defaults are used)
    domein = StripCharField(max_length=35,primary_key=True,verbose_name=_(u'Counter domain'))
    nummer = models.IntegerField(verbose_name=_(u'Last used number'))
    class Meta:
        db_table = 'uniek'
        verbose_name = _(u'counter')
        ordering = ['domein']
